#!/usr/bin/env python

"""This example demonstrates the flow for retrieving a refresh token.



This tool can be used to conveniently create refresh tokens for later use with
your web application OAuth2 credentials.

To create a Reddit application visit the following link while logged into the account
you want to create a refresh token for: https://www.reddit.com/prefs/apps/

Create a "web app" with the redirect uri set to: http://localhost:8080

After the application is created, take note of:

- REDDIT_CLIENT_ID; the line just under "web app" in the upper left of the Reddit
  Application
- REDDIT_CLIENT_SECRET; the value to the right of "secret"

Usage:

    EXPORT praw_client_id=<REDDIT_CLIENT_ID>
    EXPORT praw_client_secret=<REDDIT_CLIENT_SECRET>
    python3 obtain_refresh_token.py

"""
import asyncio
import random
import socket
import sys

import asyncpraw


async def main():
    """Provide the program's entry point when directly executed."""
    scope_input = input(
        "Enter a comma separated list of scopes, or '*' for all scopes: "
    )
    scopes = [scope.strip() for scope in scope_input.strip().split(",")]

    reddit = asyncpraw.Reddit(
        redirect_uri="http://localhost:8080",
        user_agent="obtain_refresh_token/v0 by u/bboe",
    )
    state = str(random.randint(0, 65000))
    url = reddit.auth.url(duration="permanent", scopes=scopes, state=state)
    print(f"Now open this url in your browser: {url}")

    client = receive_connection()
    data = client.recv(1024).decode("utf-8")
    param_tokens = data.split(" ", 2)[1].split("?", 1)[1].split("&")
    params = {
        key: value for (key, value) in [token.split("=") for token in param_tokens]
    }

    if state != params["state"]:
        send_message(
            client,
            f"State mismatch. Expected: {state} Received: {params['state']}",
        )
        return 1
    elif "error" in params:
        send_message(client, params["error"])
        return 1

    refresh_token = await reddit.auth.authorize(params["code"])
    send_message(client, f"Refresh token: {refresh_token}")
    await reddit._http.close()
    return 0


def receive_connection():
    """Wait for and then return a connected socket..

    Opens a TCP connection on port 8080, and waits for a single client.

    """
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(("localhost", 8080))
    server.listen(1)
    client = server.accept()[0]
    server.close()
    return client


def send_message(client, message):
    """Send message to client and close the connection."""
    print(message)
    client.send(f"HTTP/1.1 200 OK\r\n\r\n{message}".encode("utf-8"))
    client.close()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    sys.exit(loop.run_until_complete(main()))



"""Provide the DraftList class."""

from .base import BaseList


class DraftList(BaseList):
    """A list of :class:`.Draft` objects. Works just like a regular list."""

    CHILD_ATTRIBUTE = "drafts"



"""Test asyncpraw.models.ModNote."""
from asyncpraw.models.mod_note import ModNote

from .. import UnitTest


class TestModNote(UnitTest):
    def test_equality(self):
        mod_note1 = ModNote(None, {"id": "a"})
        mod_note2 = ModNote(None, {"id": "b"})
        mod_note3 = ModNote(None, {"id": "a"})
        assert mod_note1 != mod_note2
        assert mod_note1 == mod_note3
        assert mod_note1 == "a"
        assert mod_note1 != 1



from json import dumps

from pytest import raises

from asyncpraw.models import (
    Subreddit,
    SubredditWidgets,
    SubredditWidgetsModeration,
    Widget,
    WidgetModeration,
)
from asyncpraw.models.base import AsyncPRAWBase
from asyncpraw.models.reddit.widgets import WidgetEncoder

from ... import UnitTest


class TestWidgetEncoder(UnitTest):
    def test_bad_encode(self, reddit):
        data = [
            1,
            "two",
            SubredditWidgetsModeration(
                Subreddit(reddit, display_name="subreddit"), reddit
            ),
        ]
        with raises(TypeError):
            dumps(data, cls=WidgetEncoder)  # should throw TypeError

    def test_good_encode(self, reddit):
        data = [
            1,
            "two",
            AsyncPRAWBase(reddit, _data={"_secret": "no", "3": 3}),
            Subreddit(reddit, "four"),
        ]
        assert '[1, "two", {"3": 3}, "four"]' == dumps(data, cls=WidgetEncoder)


class TestWidgets(UnitTest):
    def test_subredditwidgets_mod(self, reddit):
        sw = SubredditWidgets(Subreddit(reddit, "fake_subreddit"))
        assert isinstance(sw.mod, SubredditWidgetsModeration)

    def test_widget_mod(self, reddit):
        w = Widget(reddit, {})
        assert isinstance(w.mod, WidgetModeration)
        assert w.mod.widget == w



"""Test asyncpraw.models.Trophy"""
import sys

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from .. import IntegrationTest


class TestTrophy(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_equality(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            user = await self.reddit.user.me()
            trophies = await user.trophies()
            trophies2 = await user.trophies()
            assert trophies == trophies2



import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import ClientException
from asyncpraw.models import Draft, Subreddit

from ... import IntegrationTest


class TestDraft(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

            draft = await self.reddit.drafts.create(
                title="test", url="https://reddit.com", subreddit=subreddit
            )
            await draft.load()
            assert draft.subreddit == subreddit
            assert draft.title == "test"
            assert not hasattr(draft, "selftext")
            assert draft.url == "https://reddit.com"

            draft = await self.reddit.drafts.create(
                title="test2", selftext="", subreddit=subreddit
            )
            await draft.load()
            assert draft.subreddit == subreddit
            assert draft.selftext == ""
            assert draft.title == "test2"
            assert not hasattr(draft, "url")

            draft = await self.reddit.drafts.create(
                title="test2",
                selftext="selftext",
                subreddit=pytest.placeholders.test_subreddit,
            )
            await draft.load()
            assert draft.subreddit == subreddit
            assert draft.selftext == "selftext"
            assert draft.title == "test2"
            assert not hasattr(draft, "url")

    async def test_delete(self):
        self.reddit.read_only = False
        with self.use_cassette():
            draft = await self.reddit.drafts(
                draft_id="bc647daa-3b83-11ec-a790-7e0ed9af0c88"
            )
            await draft.delete()
            assert len(await self.reddit.drafts()) == 6

    async def test_fetch(self):
        self.reddit.read_only = False
        with self.use_cassette():
            draft = await self.reddit.drafts(
                draft_id="bc647daa-3b83-11ec-a790-7e0ed9af0c88"
            )
            assert draft.title == "title"
            with pytest.raises(ClientException):
                draft = Draft(self.reddit, id="non-existent")
                await draft._fetch()

    async def test_list(self):
        self.reddit.read_only = False
        with self.use_cassette():
            assert len(await self.reddit.drafts()) == 7

    async def test_submit(self):
        self.reddit.read_only = False
        with self.use_cassette():
            total_drafts = len(await self.reddit.drafts())

            draft = await self.reddit.drafts(
                draft_id="9a77f398-3b83-11ec-8890-fa0ad7056e98"
            )
            submission = await draft.submit()
            await submission.load()
            assert submission.title == draft.title
            assert submission.selftext == draft.selftext
            assert submission.subreddit == draft.subreddit
            remaining_drafts = len(await self.reddit.drafts())
            assert remaining_drafts < total_drafts

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submit__different_subreddit(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            total_drafts = len(await self.reddit.drafts())

            draft = await self.reddit.drafts(
                draft_id="86339940-3b8e-11ec-88a2-fa740ec7656c"
            )
            submission = await draft.submit(
                subreddit=await self.reddit.subreddit(
                    pytest.placeholders.test_subreddit
                )
            )
            await submission.load()
            assert submission.title == draft.title
            assert submission.selftext == draft.selftext
            assert submission.subreddit != draft.subreddit
            remaining_drafts = len(await self.reddit.drafts())
            assert remaining_drafts < total_drafts

            draft = await self.reddit.drafts(
                draft_id="6e9a8e92-3b8e-11ec-8f1e-5291d87604fa"
            )
            submission = await draft.submit(
                subreddit=pytest.placeholders.test_subreddit
            )
            await submission.load()
            assert submission.title == draft.title
            assert submission.selftext == draft.selftext
            assert submission.subreddit != draft.subreddit
            remaining_drafts = len(await self.reddit.drafts())
            assert remaining_drafts < total_drafts

    async def test_submit__different_title(self):
        self.reddit.read_only = False
        with self.use_cassette():
            total_drafts = len(await self.reddit.drafts())

            draft = await self.reddit.drafts(
                draft_id="d3e728f6-3b8d-11ec-a2aa-fa0ad7056e98"
            )
            new_title = "new title"
            submission = await draft.submit(title=new_title)
            await submission.load()
            assert submission.title == new_title != draft.title
            assert submission.selftext == draft.selftext
            assert submission.subreddit == draft.subreddit
            remaining_drafts = len(await self.reddit.drafts())
            assert remaining_drafts < total_drafts

    async def test_update(self):
        self.reddit.read_only = False
        with self.use_cassette():
            draft = await self.reddit.drafts(
                draft_id="98de0118-3b8c-11ec-98a3-764c49cd2e1a"
            )
            assert draft.title == "title"
            await draft.update(
                title="new title", subreddit=pytest.placeholders.test_subreddit
            )
            assert draft.title == "new title"
            assert isinstance(draft.subreddit, Subreddit)
            assert draft.subreddit == pytest.placeholders.test_subreddit



import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import ClientException
from asyncpraw.models import Emoji

from ... import IntegrationTest


class TestEmoji(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test__fetch(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            emoji = await subreddit.emoji.get_emoji("test_png")
            assert emoji.created_by.startswith("t2_")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test__fetch__invalid_emoji(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            with pytest.raises(ClientException) as excinfo:
                await subreddit.emoji.get_emoji("invalid")
            assert (
                str(excinfo.value) == f"r/{subreddit} does not have the emoji invalid"
            )
            with pytest.raises(ClientException) as excinfo2:
                await subreddit.emoji.get_emoji("Test_png")
            assert (
                str(excinfo2.value) == f"r/{subreddit} does not have the emoji Test_png"
            )

    async def test_delete(self):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            emoji = await subreddit.emoji.get_emoji("test_png")
            await emoji.delete()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            emoji = await subreddit.emoji.get_emoji("test_png")
            await emoji.update(
                mod_flair_only=False, post_flair_allowed=True, user_flair_allowed=True
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update__with_preexisting_values(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            emoji = await subreddit.emoji.get_emoji("test_png")
            await emoji.update(mod_flair_only=True)


class TestSubredditEmoji(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test__iter(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            count = 0
            async for emoji in subreddit.emoji:
                assert isinstance(emoji, Emoji)
                count += 1
            assert count > 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for extension in ["jpg", "png"]:
                emoji = await subreddit.emoji.add(
                    name=f"test_{extension}",
                    image_path=f"tests/integration/files/test.{extension}",
                )
                assert isinstance(emoji, Emoji)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_with_perms(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            for extension in ["jpg", "png"]:
                emoji = await subreddit.emoji.add(
                    name=f"test_{extension}",
                    image_path=f"tests/integration/files/test.{extension}",
                    mod_flair_only=True,
                    post_flair_allowed=True,
                    user_flair_allowed=False,
                )
                assert isinstance(emoji, Emoji)



import sys

import aiohttp
import pytest

if sys.version_info < (3, 8):
    from asynctest import CoroutineMock as AsyncMock
    from asynctest import MagicMock, mock
else:
    from unittest import mock
    from unittest.mock import AsyncMock, MagicMock

from asyncpraw.exceptions import MediaPostFailed
from asyncpraw.models import InlineGif, InlineImage, InlineVideo, Subreddit, WikiPage
from asyncpraw.models.reddit.subreddit import SubredditFlairTemplates

from ... import UnitTest


class TestSubreddit(UnitTest):
    def test_equality(self, reddit):
        subreddit1 = Subreddit(reddit, _data={"display_name": "dummy1", "n": 1})
        subreddit2 = Subreddit(reddit, _data={"display_name": "Dummy1", "n": 2})
        subreddit3 = Subreddit(reddit, _data={"display_name": "dummy3", "n": 2})
        assert subreddit1 == subreddit1
        assert subreddit2 == subreddit2
        assert subreddit3 == subreddit3
        assert subreddit1 == subreddit2
        assert subreddit2 != subreddit3
        assert subreddit1 != subreddit3
        assert "dummy1" == subreddit1
        assert subreddit2 == "dummy1"

    def test_construct_failure(self, reddit):
        message = "Either 'display_name' or '_data' must be provided."
        with pytest.raises(TypeError) as excinfo:
            Subreddit(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Subreddit(reddit, "dummy", {"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(ValueError):
            Subreddit(reddit, "")

    def test_fullname(self, reddit):
        subreddit = Subreddit(reddit, _data={"display_name": "name", "id": "dummy"})
        assert subreddit.fullname == "t5_dummy"

    def test_hash(self, reddit):
        subreddit1 = Subreddit(reddit, _data={"display_name": "dummy1", "n": 1})
        subreddit2 = Subreddit(reddit, _data={"display_name": "Dummy1", "n": 2})
        subreddit3 = Subreddit(reddit, _data={"display_name": "dummy3", "n": 2})
        assert hash(subreddit1) == hash(subreddit1)
        assert hash(subreddit2) == hash(subreddit2)
        assert hash(subreddit3) == hash(subreddit3)
        assert hash(subreddit1) == hash(subreddit2)
        assert hash(subreddit2) != hash(subreddit3)
        assert hash(subreddit1) != hash(subreddit3)

    @mock.patch(
        "asyncpraw.Reddit.post", return_value={"json": {"data": {"websocket_url": ""}}}
    )
    @mock.patch(
        "asyncpraw.models.Subreddit._upload_media",
        return_value=("fake_media_url", "fake_websocket_url"),
    )
    @mock.patch("aiohttp.client.ClientSession.ws_connect")
    async def test_invalid_media(
        self, connection_mock, _mock_upload_media, _mock_post, reddit
    ):
        reddit._core._requestor._http = aiohttp.ClientSession()
        recv_mock = MagicMock()
        recv_mock.receive_json = AsyncMock(
            return_value={"payload": {}, "type": "failed"}
        )
        context_manager = MagicMock()
        context_manager.__aenter__.return_value = recv_mock
        connection_mock.return_value = context_manager

        with pytest.raises(MediaPostFailed):
            await Subreddit(reddit, display_name="test").submit_image(
                "Test", "dummy path"
            )
        await reddit._core._requestor._http.close()

    @mock.patch("asyncpraw.models.Subreddit._read_and_post_media")
    @mock.patch(
        "asyncpraw.Reddit.post",
        return_value={
            "json": {"data": {"websocket_url": ""}},
            "args": {"action": "", "fields": []},
        },
    )
    @mock.patch("aiohttp.client.ClientSession.ws_connect")
    async def test_media_upload_500(
        self, connection_mock, _mock_post, mock_method, reddit
    ):
        from aiohttp.http_exceptions import HttpProcessingError
        from asyncprawcore.exceptions import ServerError

        response = mock.Mock()
        response.status = 201
        response.raise_for_status = mock.Mock(
            side_effect=HttpProcessingError(code=500, message="")
        )
        mock_method.return_value = response
        with pytest.raises(ServerError):
            await Subreddit(reddit, display_name="test").submit_image(
                "Test", "/dev/null"
            )

    async def test_notes_delete__invalid_args(self):
        with pytest.raises(TypeError) as excinfo:
            await Subreddit(None, "SubTestBot1").mod.notes.delete(note_id="111")
        assert excinfo.value.args[0] == (
            "Either the 'redditor' parameter must be provided or this method must be"
            " called from a Redditor instance (e.g., 'redditor.notes')."
        )

    def test_repr(self, reddit):
        subreddit = Subreddit(reddit, display_name="name")
        assert repr(subreddit) == "Subreddit(display_name='name')"

    async def test_search__params_not_modified(self, reddit):
        params = {"dummy": "value"}
        subreddit = Subreddit(reddit, display_name="name")
        generator = subreddit.search(None, params=params)
        assert generator.params["dummy"] == "value"
        assert params == {"dummy": "value"}

    def test_str(self, reddit):
        subreddit = Subreddit(reddit, _data={"display_name": "name", "id": "dummy"})
        assert str(subreddit) == "name"

    async def test_submit_failure(self, reddit):
        message = "Either 'selftext' or 'url' must be provided."
        subreddit = Subreddit(reddit, display_name="name")

        with pytest.raises(TypeError) as excinfo:
            await subreddit.submit("Cool title")
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            await subreddit.submit("Cool title", selftext="a", url="b")
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            await subreddit.submit("Cool title", selftext="", url="b")
        assert str(excinfo.value) == message

    async def test_upload_banner_additional_image(self, reddit):
        subreddit = Subreddit(reddit, display_name="name")
        with pytest.raises(ValueError):
            await subreddit.stylesheet.upload_banner_additional_image(
                "dummy_path", align="asdf"
            )

    async def test_submit_gallery__missing_path(self, reddit):
        message = "'image_path' is required."
        subreddit = Subreddit(reddit, display_name="name")

        with pytest.raises(TypeError) as excinfo:
            await subreddit.submit_gallery(
                "Cool title", [{"caption": "caption"}, {"caption": "caption2"}]
            )
        assert str(excinfo.value) == message

    async def test_submit_gallery__invalid_path(self, reddit):
        message = "'invalid_image_path' is not a valid image path."
        subreddit = Subreddit(reddit, display_name="name")

        with pytest.raises(TypeError) as excinfo:
            await subreddit.submit_gallery(
                "Cool title", [{"image_path": "invalid_image_path"}]
            )
        assert str(excinfo.value) == message

    async def test_submit_gallery__too_long_caption(self, reddit):
        message = "Caption must be 180 characters or less."
        subreddit = Subreddit(reddit, display_name="name")
        caption = (
            "wayyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy"
            "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy"
            "yyyyyyyyyyyyyyyy too long caption"
        )
        with pytest.raises(TypeError) as excinfo:
            await subreddit.submit_gallery(
                "Cool title", [{"image_path": __file__, "caption": caption}]
            )
        assert str(excinfo.value) == message

    async def test_submit_inline_media__invalid_path(self, reddit):
        message = "'invalid_image_path' is not a valid file path."
        subreddit = Subreddit(reddit, display_name="name")
        gif = InlineGif("invalid_image_path", "optional caption")
        image = InlineImage("invalid_image_path", "optional caption")
        video = InlineVideo("invalid_image_path", "optional caption")
        selftext = "Text with {gif1}, {image1}, and {video1} inline"
        media = {"gif1": gif, "image1": image, "video1": video}
        with pytest.raises(ValueError) as excinfo:
            await subreddit.submit("title", inline_media=media, selftext=selftext)
        assert str(excinfo.value) == message


class TestSubredditFlair(UnitTest):
    async def test_set(self, reddit):
        subreddit = Subreddit(reddit, pytest.placeholders.test_subreddit)
        with pytest.raises(TypeError):
            await subreddit.flair.set(
                "a_redditor", css_class="myCSS", flair_template_id="gibberish"
            )


class TestSubredditFlairTemplates(UnitTest):
    async def test_not_implemented(self, reddit):
        with pytest.raises(NotImplementedError):
            await SubredditFlairTemplates(
                Subreddit(reddit, pytest.placeholders.test_subreddit)
            ).__aiter__()


class TestSubredditWiki(UnitTest):
    async def test__getitem(self, reddit):
        subreddit = Subreddit(reddit, display_name="name")
        wikipage = await subreddit.wiki.get_page("Foo", fetch=False)
        assert isinstance(wikipage, WikiPage)
        assert "foo" == wikipage.name


class TestSubredditModmailConversationsStream(UnitTest):
    async def test_conversation_stream_init(self, reddit):
        submodstream = Subreddit(reddit, display_name="mod").mod.stream
        submodstream.modmail_conversations()
        assert submodstream.subreddit == "all"

    async def test_conversation_stream_capitalization(self, reddit):
        submodstream = Subreddit(reddit, display_name="Mod").mod.stream
        submodstream.modmail_conversations()
        assert submodstream.subreddit == "all"



"""Test asyncpraw.models.redditor."""
import sys

import pytest
from asyncprawcore import Forbidden

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import RedditAPIException
from asyncpraw.models import Comment, Redditor, Submission

from ... import IntegrationTest


class TestRedditor(IntegrationTest):
    FRIEND = "PyAPITestUser3"
    FRIEND_FULLNAME = "t2_6c1xj"

    async def test_block(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(self.FRIEND, fetch=True)
            await redditor.block()

    async def test_friend(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(self.FRIEND.lower(), fetch=True)
            await redditor.friend()

    async def test_friend__with_note__no_gold(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                redditor = await self.reddit.redditor(self.FRIEND.lower(), fetch=True)
                await redditor.friend(note="asyncpraw")
            assert "GOLD_REQUIRED" == excinfo.value.error_type

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_friend_info(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            friend = await self.reddit.redditor(self.FRIEND, fetch=True)
            redditor = await friend.friend_info()
            assert self.FRIEND == redditor
            assert "date" in redditor.__dict__
            assert "created_utc" not in redditor.__dict__
            await redditor.load()
            assert hasattr(redditor, "created_utc")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_fullname_init(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(
                fetch=True, fullname=self.FRIEND_FULLNAME
            )
            assert redditor.name == self.FRIEND

    async def test_gild__no_creddits(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                redditor = await self.reddit.redditor("subreddit_stats")
                await redditor.gild()
            assert "INSUFFICIENT_COINS" == excinfo.value.error_type

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_message(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("subreddit_stats")
            await redditor.message(
                subject="Async PRAW test", message="This is a test from Async PRAW"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_message_from_subreddit(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("subreddit_stats")
            await redditor.message(
                subject="Async PRAW test",
                message="This is a test from Async PRAW",
                from_subreddit=pytest.placeholders.test_subreddit,
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderated(self, _):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            redditor_no_mod = await self.reddit.redditor("Turambar420")
            moderated = await redditor.moderated()
            assert len(moderated) > 0
            assert len(moderated[0].display_name) > 0
            not_moderated = await redditor_no_mod.moderated()
            assert len(not_moderated) == 0

    async def test_multireddits(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("kjoneslol")
            multireddits = await redditor.multireddits()
            for multireddit in multireddits:
                if "sfwpornnetwork" == multireddit.name:
                    break
            else:
                assert False, "sfwpornnetwork not found in multireddits"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_stream__comments(self, _):
        with self.use_cassette():
            redditor = await self.reddit.redditor("AutoModerator")
            generator = redditor.stream.comments()
            for i in range(101):
                assert isinstance(await self.async_next(generator), Comment)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_stream__submissions(self, _):
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            generator = redditor.stream.submissions()
            for i in range(101):
                assert isinstance(await self.async_next(generator), Submission)

    async def test_trophies(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            trophies = await redditor.trophies()
            assert len(trophies) > 0
            assert len(trophies[0].name) > 0

    async def test_trophies__user_not_exist(self):
        with self.use_cassette():
            redditor = Redditor(self.reddit, "thisusershouldnotexist")
            with pytest.raises(RedditAPIException) as excinfo:
                await redditor.trophies()
            assert "USER_DOESNT_EXIST" == excinfo.value.error_type

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unblock(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            blocked = await self.reddit.user.blocked()
            for redditor in blocked:
                await redditor.unblock()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unfriend(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            friends = await self.reddit.user.friends()
            redditor = friends[0]
            assert await redditor.unfriend() is None

    async def test_notes__subreddits(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("Lil_SpazTest")
            notes = await self.async_list(
                redditor.notes.subreddits(
                    pytest.placeholders.test_subreddit, "Lil_SpazTest"
                )
            )
            assert len(notes) == 2
            assert notes[0].user == redditor
            assert notes[1] is None


class TestRedditorListings(IntegrationTest):
    async def test_comments__controversial(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            comments = await self.async_list(redditor.comments.controversial())
        assert len(comments) == 100

    async def test_comments__hot(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            comments = await self.async_list(redditor.comments.hot())
        assert len(comments) == 100

    async def test_comments__new(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            comments = await self.async_list(redditor.comments.new())
        assert len(comments) == 100

    async def test_comments__top(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            comments = await self.async_list(redditor.comments.top())
        assert len(comments) == 100

    async def test_controversial(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            items = await self.async_list(redditor.controversial())
        assert len(items) == 100

    async def test_downvoted(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            submissions = await self.async_list(redditor.downvoted())
        assert len(submissions) > 0

    async def test_downvoted__in_read_only_mode(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            with pytest.raises(Forbidden):
                await self.async_list(redditor.downvoted())

    async def test_downvoted__other_user(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            with pytest.raises(Forbidden):
                await self.async_list(redditor.downvoted())

    async def test_gilded(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            items = await self.async_list(redditor.gilded(limit=50))
        assert len(items) == 50

    async def test_gildings(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            items = await self.async_list(redditor.gildings())
        assert isinstance(items, list)

    async def test_gildings__in_read_only_mode(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            with pytest.raises(Forbidden):
                await self.async_list(redditor.gildings())

    async def test_gildings__other_user(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            with pytest.raises(Forbidden):
                await self.async_list(redditor.gildings())

    async def test_hidden(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            submissions = await self.async_list(redditor.hidden())
        assert len(submissions) > 0

    async def test_hidden__in_read_only_mode(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            with pytest.raises(Forbidden):
                await self.async_list(redditor.hidden())

    async def test_hidden__other_user(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            with pytest.raises(Forbidden):
                await self.async_list(redditor.hidden())

    async def test_hot(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            items = await self.async_list(redditor.hot())
        assert len(items) == 100

    async def test_new(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            items = await self.async_list(redditor.new())
        assert len(items) == 100

    async def test_saved(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            items = await self.async_list(redditor.saved())
        assert len(items) > 0

    async def test_saved__in_read_only_mode(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            with pytest.raises(Forbidden):
                await self.async_list(redditor.saved())

    async def test_saved__other_user(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            with pytest.raises(Forbidden):
                await self.async_list(redditor.saved())

    async def test_submissions__controversial(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spladug")
            submissions = await self.async_list(redditor.submissions.controversial())
        assert len(submissions) == 100

    async def test_submissions__hot(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            submissions = await self.async_list(redditor.submissions.hot())
        assert len(submissions) == 100

    async def test_submissions__new(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            submissions = await self.async_list(redditor.submissions.new())
        assert len(submissions) == 100

    async def test_submissions__top(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spladug")
            submissions = await self.async_list(redditor.submissions.top())
        assert len(submissions) == 100

    async def test_top(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            items = await self.async_list(redditor.top())
        assert len(items) == 100

    async def test_trust_and_distrust(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("PyAPITestUser3")
            await redditor.trust()
            redditor = (await self.reddit.user.trusted())[0]
            await redditor.distrust()

    async def test_trust_blocked_user(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("kn0thing")
            await redditor.block()
            with pytest.raises(RedditAPIException) as excinfo:
                await redditor.trust()
            assert "CANT_WHITELIST_AN_ENEMY" == excinfo.value.error_type

    async def test_upvoted(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            submissions = await self.async_list(redditor.upvoted())
        assert len(submissions) > 0

    async def test_upvoted__in_read_only_mode(self):
        with self.use_cassette():
            redditor = await self.reddit.redditor(pytest.placeholders.username)
            with pytest.raises(Forbidden):
                await self.async_list(redditor.upvoted())

    async def test_upvoted__other_user(self):
        self.reddit.read_only = False
        with self.use_cassette():
            redditor = await self.reddit.redditor("spez")
            with pytest.raises(Forbidden):
                await self.async_list(redditor.upvoted())



"""Async PRAW Test Suite."""
import sys

if sys.version_info < (3, 8):
    from asynctest import TestCase

    class BaseTest(TestCase):
        """Base class for Async PRAW tests."""

else:

    class BaseTest:
        """Base class for Async PRAW tests."""


class HelperMethodMixin:
    @staticmethod
    async def async_list(async_generator):
        """Return a list from an async iterator."""
        return [item async for item in async_generator]

    @staticmethod
    async def async_next(async_generator):
        """Return the next item from an async iterator."""
        return await async_generator.__anext__()



"""Provide the :class:`.UserSubreddit` class."""
import inspect
from typing import TYPE_CHECKING, Dict, Union
from warnings import warn

from ...util.cache import cachedproperty
from .subreddit import Subreddit, SubredditModeration

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class UserSubreddit(Subreddit):
    """A class for :class:`.User` Subreddits.

    To obtain an instance of this class execute:

    .. code-block:: python

        subreddit = (await reddit.user.me()).subreddit

    .. include:: ../../typical_attributes.rst

    ========================= ==========================================================
    Attribute                 Description
    ========================= ==========================================================
    ``can_assign_link_flair`` Whether users can assign their own link flair.
    ``can_assign_user_flair`` Whether users can assign their own user flair.
    ``created_utc``           Time the subreddit was created, represented in `Unix
                              Time`_.
    ``description``           Subreddit description, in Markdown.
    ``description_html``      Subreddit description, in HTML.
    ``display_name``          Name of the subreddit.
    ``icon_img``              The URL of the subreddit icon image.
    ``id``                    ID of the subreddit.
    ``name``                  Fullname of the subreddit.
    ``over18``                Whether the subreddit is NSFW.
    ``public_description``    Description of the subreddit, shown in searches and on the
                              "You must be invited to visit this community" page (if
                              applicable).
    ``spoilers_enabled``      Whether the spoiler tag feature is enabled.
    ``subscribers``           Count of subscribers. This will be ``0`` unless unless the
                              authenticated user is a moderator.
    ``user_is_banned``        Whether the authenticated user is banned.
    ``user_is_moderator``     Whether the authenticated user is a moderator.
    ``user_is_subscriber``    Whether the authenticated user is subscribed.
    ========================= ==========================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    def __init__(self, reddit: "asyncpraw.Reddit", *args, **kwargs):
        """Initialize an :class:`.UserSubreddit` instance.

        :param reddit: An instance of :class:`.Reddit`.

        .. note::

            This class should not be initialized directly. Instead, obtain an instance
            via:

            .. code-block:: python

                user = await reddit.user.me()
                user.subreddit

                # or

                redditor = await reddit.redditor("redditor_name")
                redditor.subreddit

        """

        def predicate(item):
            name = getattr(item, "__name__", None)
            return name not in dir(object) + dir(Subreddit) and name in dir(dict)

        for name, member in inspect.getmembers(dict, predicate=predicate):
            if name != "__getitem__":
                setattr(
                    self,
                    name,
                    self._dict_depreciated_wrapper(getattr(self.__dict__, name)),
                )

        super().__init__(reddit, *args, **kwargs)

    @staticmethod
    def _dict_depreciated_wrapper(func):
        """Show deprecation notice for dict only methods."""

        def wrapper(*args, **kwargs):
            warn(
                "'Redditor.subreddit' is no longer a dict and is now an UserSubreddit"
                f" object. Using '{func.__name__}' is deprecated and will be removed in"
                " Async PRAW 8.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapper

    @cachedproperty
    def mod(self) -> "asyncpraw.models.reddit.user_subreddit.UserSubredditModeration":
        """Provide an instance of :class:`.UserSubredditModeration`.

        For example, to update the authenticated user's display name:

        .. code-block:: python

            user = await reddit.user.me()
            await user.subreddit.mod.update(title="New display name")

        """
        return UserSubredditModeration(self)

    def __getitem__(self, item):
        """Show deprecation notice for dict method ``__getitem__``."""
        warn(
            "'Redditor.subreddit' is no longer a dict and is now an UserSubreddit"
            " object. Accessing attributes using string indices is deprecated.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return getattr(self, item)


class UserSubredditModeration(SubredditModeration):
    """Provides a set of moderation functions to a :class:`.UserSubreddit`.

    For example, to accept a moderation invite from the user subreddit of u/spez:

    .. code-block:: python

        redditor = await reddit.redditor("spez")
        await redditor.subreddit.mod.accept_invite()

    """

    async def update(
        self, **settings: Union[str, int, bool]
    ) -> Dict[str, Union[str, int, bool]]:
        """Update the :class:`.Subreddit`'s settings.

        :param all_original_content: Mandate all submissions to be original content
            only.
        :param allow_chat_post_creation: Allow users to create chat submissions.
        :param allow_images: Allow users to upload images using the native image
            hosting.
        :param allow_polls: Allow users to post polls to the subreddit.
        :param allow_post_crossposts: Allow users to crosspost submissions from other
            subreddits.
        :param allow_top: Allow the subreddit to appear on r/all as well as the default
            and trending lists.
        :param allow_videos: Allow users to upload videos using the native image
            hosting.
        :param collapse_deleted_comments: Collapse deleted and removed comments on
            comments pages by default.
        :param crowd_control_chat_level: Controls the crowd control level for chat
            rooms. Goes from 0-3.
        :param crowd_control_level: Controls the crowd control level for submissions.
            Goes from 0-3.
        :param crowd_control_mode: Enables/disables crowd control.
        :param comment_score_hide_mins: The number of minutes to hide comment scores.
        :param description: Shown in the sidebar of your subreddit.
        :param disable_contributor_requests: Specifies whether redditors may send
            automated modmail messages requesting approval as a submitter.
        :param exclude_banned_modqueue: Exclude posts by site-wide banned users from
            modqueue/unmoderated.
        :param free_form_reports: Allow users to specify custom reasons in the report
            menu.
        :param header_hover_text: The text seen when hovering over the snoo.
        :param hide_ads: Don't show ads within this subreddit. Only applies to
            Premium-user only subreddits.
        :param key_color: A 6-digit rgb hex color (e.g., ``"#AABBCC"``), used as a
            thematic color for your subreddit on mobile.
        :param lang: A valid IETF language tag (underscore separated).
        :param link_type: The types of submissions users can make. One of ``"any"``,
            ``"link"``, or ``"self"``.
        :param original_content_tag_enabled: Enables the use of the ``original content``
            label for submissions.
        :param over_18: Viewers must be over 18 years old (i.e., NSFW).
        :param public_description: Public description blurb. Appears in search results
            and on the landing page for private subreddits.
        :param public_traffic: Make the traffic stats page public.
        :param restrict_commenting: Specifies whether approved users have the ability to
            comment.
        :param restrict_posting: Specifies whether approved users have the ability to
            submit posts.
        :param show_media: Show thumbnails on submissions.
        :param show_media_preview: Expand media previews on comments pages.
        :param spam_comments: Spam filter strength for comments. One of ``"all"``,
            ``"low"``, or ``"high"``.
        :param spam_links: Spam filter strength for links. One of ``"all"``, ``"low"``,
            or ``"high"``.
        :param spam_selfposts: Spam filter strength for selfposts. One of ``"all"``,
            ``"low"``, or ``"high"``.
        :param spoilers_enabled: Enable marking posts as containing spoilers.
        :param submit_link_label: Custom label for submit link button (None for
            default).
        :param submit_text: Text to show on submission page.
        :param submit_text_label: Custom label for submit text post button (None for
            default).
        :param subreddit_type: The string ``"user"``.
        :param suggested_comment_sort: All comment threads will use this sorting method
            by default. Leave ``None``, or choose one of ``confidence``,
            ``"controversial"``, ``"live"``, ``"new"``, ``"old"``, ``"qa"``,
            ``"random"``, or ``"top"``.
        :param title: The title of the subreddit.
        :param welcome_message_enabled: Enables the subreddit welcome message.
        :param welcome_message_text: The text to be used as a welcome message. A welcome
            message is sent to all new subscribers by a Reddit bot.
        :param wiki_edit_age: Account age, in days, required to edit and create wiki
            pages.
        :param wiki_edit_karma: Subreddit karma required to edit and create wiki pages.
        :param wikimode: One of ``"anyone"``, ``"disabled"``, or ``"modonly"``.

        Additional keyword arguments can be provided to handle new settings as Reddit
        introduces them.

        Settings that are documented here and aren't explicitly set by you in a call to
        :meth:`.SubredditModeration.update` should retain their current value. If they
        do not please file a bug.

        .. warning::

            Undocumented settings, or settings that were very recently documented, may
            not retain their current value when updating. This often occurs when Reddit
            adds a new setting but forgets to add that setting to the API endpoint that
            is used to fetch the current settings.

        """
        current_settings = await self.settings()

        # These attributes come out using different names than they go in.
        remap = {
            "allow_top": "default_set",
            "header_title": "header_hover_text",
            "lang": "language",
            "link_type": "content_options",
            "sr": "subreddit_id",
            "type": "subreddit_type",
        }
        for new, old in remap.items():
            current_settings[new] = current_settings.pop(old)

        current_settings.update(settings)
        return await UserSubreddit._create_or_update(
            _reddit=self.subreddit._reddit, **current_settings
        )



import os
import sys
from datetime import datetime

# Do not touch these. They use the local Async PRAW over the global Async PRAW.
sys.path.insert(0, ".")
sys.path.insert(1, "..")

from asyncpraw import __version__  # noqa: E402

copyright = datetime.today().strftime("%Y, Joel Payne")
exclude_patterns = ["_build"]
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_dark_mode",
    "sphinx_rtd_theme",
    "sphinxcontrib_trio",
]
html_static_path = ["_static"]
html_theme = "sphinx_rtd_theme"
html_theme_options = {"collapse_navigation": True}
htmlhelp_basename = "Async PRAW"
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}
master_doc = "index"
nitpick_ignore = [
    ("py:class", "IO"),
    ("py:class", "asyncprawcore.requestor.Requestor"),
    ("py:class", "asyncpraw.models.redditors.PartialRedditor"),
]
nitpicky = True
project = "Async PRAW"
pygments_style = "sphinx"
release = __version__
source_suffix = ".rst"
suppress_warnings = ["image.nonlocal_uri"]
version = ".".join(__version__.split(".", 2)[:2])


# Use RTD theme locally
if not os.environ.get("READTHEDOCS"):
    import sphinx_rtd_theme

    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


def skip(app, what, name, obj, skip, options):
    if name in {
        "__call__",
        "__contains__",
        "__getitem__",
        "__init__",
        "__iter__",
        "__len__",
    }:
        return False
    return skip


def setup(app):
    app.connect("autodoc-skip-member", skip)
    app.add_css_file("theme_override.css")



"""Test asyncpraw.models.front."""
import pytest

from .. import UnitTest


class TestFront(UnitTest):
    def test_controversial_raises_value_error(self, reddit):
        with pytest.raises(ValueError):
            reddit.front.controversial(time_filter="second")

    def test_top_raises_value_error(self, reddit):
        with pytest.raises(ValueError):
            reddit.front.top(time_filter="second")



import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import AsyncPRAWException, ClientException, RedditAPIException
from asyncpraw.models import Comment, Submission

from ... import IntegrationTest


class TestComment(IntegrationTest):
    async def test_attributes(self):
        with self.use_cassette():
            comment = await self.reddit.comment("cklhv0f")
            assert comment.author == "bboe"
            assert comment.body.startswith("Yes it does.")
            assert not comment.is_root
            assert comment.submission == "2gmzqe"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_block(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = None
            async for item in self.reddit.inbox.submission_replies():
                if item.author and item.author != pytest.placeholders.username:
                    comment = item
                    break
            else:
                assert False, "no comment found"
            await comment.block()

    async def test_clear_vote(self):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.reddit.comment("fwxxs5d")
            await comment.clear_vote()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = Comment(self.reddit, "fx1tgzm")
            await comment.delete()
            comment = await self.reddit.comment("fx1tgzm")
            assert comment.author is None
            assert comment.body == "[deleted]"

    async def test_disable_inbox_replies(self):
        self.reddit.read_only = False
        comment = Comment(self.reddit, "fwxxs5d")
        with self.use_cassette():
            await comment.disable_inbox_replies()

    async def test_downvote(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fwxxs5d").downvote()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_edit(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = Comment(self.reddit, "fwxxs5d")
            await comment.edit("New text")
            assert comment.body == "New text"

    async def test_enable_inbox_replies(self):
        self.reddit.read_only = False
        comment = Comment(self.reddit, "fwxxs5d")
        with self.use_cassette():
            await comment.enable_inbox_replies()

    async def test_award(self):
        self.reddit.read_only = False
        with self.use_cassette():
            award_data = await Comment(self.reddit, "g7cmlgc").award()
            assert award_data["gildings"]["gid_2"] == 2

    async def test_award__not_enough_coins(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await Comment(self.reddit, "g7cmlgc").award(
                    gild_type="award_2385c499-a1fb-44ec-b9b7-d260f3dc55de"
                )
            exception = excinfo.value
            assert "INSUFFICIENT_COINS_WITH_AMOUNT" == exception.error_type

    async def test_award__self_gild(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await Comment(self.reddit, "g7cn9xb").award(
                    gild_type="award_2385c499-a1fb-44ec-b9b7-d260f3dc55de"
                )
            exception = excinfo.value
            assert "SELF_GILDING_NOT_ALLOWED" == exception.error_type

    async def test_gild(self):
        self.reddit.read_only = False
        with self.use_cassette("TestComment.test_award"):
            award_data = await Comment(self.reddit, "g7cmlgc").gild()
            assert award_data["gildings"]["gid_2"] == 2

    async def test_invalid(self):
        with self.use_cassette():
            with pytest.raises(AsyncPRAWException) as excinfo:
                await self.reddit.comment("0")
            assert excinfo.value.args[0].startswith("No data returned for comment")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mark_read(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.async_next(self.reddit.inbox.unread())
            assert isinstance(comment, Comment)
            await comment.mark_read()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mark_unread(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.async_next(self.reddit.inbox.comment_replies())
            await comment.mark_unread()

    async def test_notes(self):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.reddit.comment("id4adsw")
            note = await comment.mod.create_note(label="HELPFUL_USER", note="test note")
            notes = await self.async_list(comment.mod.author_notes())
            assert note in notes
            assert notes[notes.index(note)].user == comment.author

    async def test_parent__comment(self):
        comment = Comment(self.reddit, "cklhv0f")
        with self.use_cassette():
            parent = await comment.parent()
            await parent.refresh()
            assert comment in parent.replies
        assert isinstance(parent, Comment)
        assert parent.fullname == comment.parent_id

    async def test_parent__chain(self):
        comment = Comment(self.reddit, "dkk4qjd")
        counter = 0
        with self.use_cassette():
            await comment.refresh()
            parent = await comment.parent()
            while parent != comment.submission:
                if counter % 9 == 0:
                    await parent.refresh()
                counter += 1
                parent = await parent.parent()

    async def test_parent__comment_from_forest(self):
        with self.use_cassette():
            submission = await self.reddit.submission("2gmzqe")
            comment = submission.comments[0].replies[0]
        parent = await comment.parent()
        assert comment in parent.replies
        assert isinstance(parent, Comment)
        assert parent.fullname == comment.parent_id

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_parent__from_replies(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.async_next(self.reddit.inbox.comment_replies())
            parent = await comment.parent()
        assert isinstance(parent, Comment)
        assert parent.fullname == comment.parent_id

    async def test_parent__submission(self):
        comment = Comment(self.reddit, "cklfmye")
        with self.use_cassette():
            parent = await comment.parent()
            await parent.load()
            assert comment in parent.comments
        assert isinstance(parent, Submission)
        assert parent.fullname == comment.parent_id

    async def test_refresh(self):
        with self.use_cassette():
            comment = await self.reddit.comment("d81vwef")
            assert len(comment.replies) == 0
            await comment.refresh()
            assert len(comment.replies) > 0

    async def test_refresh__raises_exception(self):
        with self.use_cassette():
            with pytest.raises(ClientException) as excinfo:
                await Comment(self.reddit, "fx1tgzm").refresh()
        assert (
            "This comment does not appear to be in the comment tree",
        ) == excinfo.value.args

    async def test_refresh__twice(self):
        with self.use_cassette():
            comment = await Comment(self.reddit, "d81vwef").refresh()
            await comment.refresh()

    async def test_refresh__deleted_comment(self):
        with self.use_cassette():
            with pytest.raises(ClientException) as excinfo:
                await Comment(self.reddit, "d7ltvl0").refresh()
        assert (
            "This comment does not appear to be in the comment tree",
        ) == excinfo.value.args

    async def test_refresh__removed_comment(self):
        with self.use_cassette():
            with pytest.raises(ClientException) as excinfo:
                await Comment(self.reddit, "fx1hmwb").refresh()
        assert (
            "This comment does not appear to be in the comment tree",
        ) == excinfo.value.args

    async def test_refresh__with_reply_sort_and_limit(self):
        with self.use_cassette():
            comment = Comment(self.reddit, "e4j4830")
            comment.reply_limit = 4
            comment.reply_sort = "new"
            await comment.refresh()
            replies = comment.replies
        last_created = float("inf")
        for reply in replies:
            if isinstance(reply, Comment):
                if reply.created_utc > last_created:
                    assert False, "sort order incorrect"
                last_created = reply.created_utc
        assert len(comment.replies) == 3

    async def test_reply(self):
        self.reddit.read_only = False
        with self.use_cassette():
            parent_comment = Comment(self.reddit, "fx1ec2p")
            comment = await parent_comment.reply("Comment reply")
            assert comment.author == pytest.placeholders.username
            assert comment.body == "Comment reply"
            assert not comment.is_root
            assert comment.parent_id == parent_comment.fullname

    # async def test_reply__none(self): # TODO: I have not been able to reproduce this again; same with praw
    #     self.reddit.read_only = False
    #     comment = Comment(self.reddit, "fx1rxr1")
    #     with self.use_cassette():
    #         reply = await comment.reply("TEST")
    #     assert reply is None

    async def test_report(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1it87").report("custom")

    async def test_save(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx19hsi").save(category="foo")

    async def test_unsave(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx19hsi").unsave()

    async def test_upvote(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx19hsi").upvote()


class TestCommentModeration(IntegrationTest):
    async def test_approve(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.approve()

    async def test_distinguish(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fy79crc").mod.distinguish()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_distinguish__sticky(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fy79crc").mod.distinguish(sticky=True)

    async def test_ignore_reports(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.ignore_reports()

    async def test_lock(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.lock()

    async def test_remove(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.remove(spam=True)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_remove_with_reason_id(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.remove(reason_id="157l6k4g6s365")

    async def test_show(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.show()

    async def test_unlock(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.unlock()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_removal_reason(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.reddit.comment("fx1jgsp")
            await comment.mod.remove()
            await comment.mod._add_removal_reason(
                mod_note="Blah", reason_id="157l6k4g6s365"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_removal_reason_without_id(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.reddit.comment("fx1jgsp")
            await comment.mod.remove()
            await comment.mod._add_removal_reason(mod_note="Test")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_removal_reason_without_id_or_note(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(ValueError) as excinfo:
                comment = await self.reddit.comment("fx1jgsp")
                await comment.mod.remove()
                await comment.mod._add_removal_reason()
            assert excinfo.value.args[0].startswith("mod_note cannot be blank")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_send_removal_message(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.reddit.comment("fx1jgsp")
            mod = comment.mod
            await mod.remove()
            message = "message"
            res = [
                await mod.send_removal_message(
                    message=message, title="title", type=type
                )
                for type in ("public", "private", "private_exposed")
            ]
            assert isinstance(res[0], Comment)
            assert res[0].parent_id == f"t1_{comment.id}"
            assert res[0].body == message
            assert res[1] is None
            assert res[2] is None

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_send_removal_message__error(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.reddit.comment("fx1jgsp")
            await comment.mod.remove()
            with pytest.raises(RedditAPIException) as excinfo:
                await comment.mod.send_removal_message(
                    message="message", title="a" * 51
                )
            exception = excinfo.value
            assert "title" == exception.field
            assert "TOO_LONG" == exception.error_type

    async def test_undistinguish(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fy79crc").mod.undistinguish()

    async def test_unignore_reports(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Comment(self.reddit, "fx1jgsp").mod.unignore_reports()



"""Provide the FullnameMixin class."""


class FullnameMixin:
    """Interface for classes that have a fullname."""

    _kind = None

    @property
    def fullname(self) -> str:
        """Return the object's fullname.

        A fullname is an object's kind mapping like ``t3`` followed by an underscore and
        the object's base36 ID, e.g., ``t1_c5s96e0``.

        """
        if "_" in self.id:
            return self.id
        return f"{self._kind}_{self.id}"



import configparser
import sys
import types

import pytest
from asyncprawcore import Requestor
from asyncprawcore.exceptions import BadRequest

if sys.version_info < (3, 8):
    from asynctest import CoroutineMock as AsyncMock
    from asynctest import mock
else:
    from unittest import mock
    from unittest.mock import AsyncMock

from asyncpraw import Reddit, __version__
from asyncpraw.config import Config
from asyncpraw.exceptions import ClientException, RedditAPIException
from asyncpraw.models import Comment
from asyncpraw.util.token_manager import BaseTokenManager

from . import UnitTest


class DummyTokenManager(BaseTokenManager):
    def post_refresh_callback(self, authorizer):
        pass

    def pre_refresh_callback(self, authorizer):
        pass


class TestReddit(UnitTest):
    REQUIRED_DUMMY_SETTINGS = {
        x: "dummy" for x in ["client_id", "client_secret", "user_agent"]
    }

    async def test_close_session(self):
        temp_reddit = Reddit(**self.REQUIRED_DUMMY_SETTINGS)
        assert not temp_reddit.requestor._http.closed
        async with temp_reddit as reddit:
            pass
        assert reddit.requestor._http.closed and temp_reddit.requestor._http.closed

    @mock.patch("asyncpraw.reddit.update_check", create=True)
    @mock.patch("asyncpraw.reddit.UPDATE_CHECKER_MISSING", False)
    @mock.patch("asyncpraw.reddit.Reddit.update_checked", False)
    async def test_check_for_updates(self, mock_update_check):
        Reddit(check_for_updates="1", **self.REQUIRED_DUMMY_SETTINGS)
        assert Reddit.update_checked
        mock_update_check.assert_called_with("asyncpraw", __version__)

    @mock.patch("asyncpraw.reddit.update_check", create=True)
    @mock.patch("asyncpraw.reddit.UPDATE_CHECKER_MISSING", True)
    @mock.patch("asyncpraw.reddit.Reddit.update_checked", False)
    async def test_check_for_updates_update_checker_missing(self, mock_update_check):
        Reddit(check_for_updates="1", **self.REQUIRED_DUMMY_SETTINGS)
        assert not Reddit.update_checked
        assert not mock_update_check.called

    def test_comment(self, reddit):
        assert Comment(reddit, id="cklfmye").id == "cklfmye"

    def test_conflicting_settings(self):
        with pytest.raises(TypeError) as excinfo:
            Reddit(
                token_manager="dummy",
                refresh_token="dummy",
                **self.REQUIRED_DUMMY_SETTINGS,
            )
        assert (
            str(excinfo.value)
            == "'refresh_token' setting cannot be provided when providing"
            " 'token_manager'"
        )

    async def test_context_manager(self):
        async with Reddit(**self.REQUIRED_DUMMY_SETTINGS) as reddit:
            assert not reddit._validate_on_submit
            assert not reddit.requestor._http.closed
        assert reddit.requestor._http.closed

    def test_info__invalid_param(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            reddit.info(fullnames=None)

        err_str = "Either 'fullnames', 'url', or 'subreddits' must be provided."
        assert str(excinfo.value) == err_str

        with pytest.raises(TypeError) as excinfo:
            reddit.info(fullnames=[], url="")

        assert str(excinfo.value) == err_str

    def test_invalid_config(self):
        with pytest.raises(ValueError) as excinfo:
            Reddit(timeout="test", **self.REQUIRED_DUMMY_SETTINGS)
        assert (
            excinfo.value.args[0]
            == "An incorrect config type was given for option timeout. The expected"
            " type is int, but the given value is test."
        )
        with pytest.raises(ValueError) as excinfo:
            Reddit(ratelimit_seconds="test", **self.REQUIRED_DUMMY_SETTINGS)
        assert (
            excinfo.value.args[0]
            == "An incorrect config type was given for option ratelimit_seconds. The"
            " expected type is int, but the given value is test."
        )

    def test_info__not_list(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            reddit.info(fullnames="Let's try a string")

        assert "must be a non-str iterable" in str(excinfo.value)

    def test_live_info__valid_param(self, reddit):
        gen = reddit.live.info(["dummy", "dummy2"])
        assert isinstance(gen, types.AsyncGeneratorType)

    def test_live_info__invalid_param(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            reddit.live.info(None)
        assert str(excinfo.value) == "ids must be a list"

    async def test_multireddit(self, reddit):
        multireddit = await reddit.multireddit(redditor="bboe", name="aa")
        assert multireddit.path == "/user/bboe/m/aa"

    @mock.patch(
        "asyncpraw.Reddit.request",
        side_effect=[
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "Some unexpected error message",
                            "ratelimit",
                        ]
                    ]
                }
            },
        ],
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_ratelimit__invalid_rate_limit_message(
        self, mock_sleep, _, reddit
    ):
        with pytest.raises(RedditAPIException) as exception:
            await reddit.post("test")
        assert exception.value.message == "Some unexpected error message"
        mock_sleep.assert_not_called()

    @mock.patch(
        "asyncpraw.Reddit.request",
        side_effect=[
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 6 seconds.",
                            "ratelimit",
                        ]
                    ]
                }
            },
        ],
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_ratelimit__over_threshold__seconds(self, mock_sleep, _, reddit):
        with pytest.raises(RedditAPIException) as exception:
            await reddit.post("test")
        assert (
            exception.value.message
            == "You are doing that too much. Try again in 6 seconds."
        )
        mock_sleep.assert_not_called()

    @mock.patch(
        "asyncpraw.Reddit.request",
        side_effect=[
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 1 minute.",
                            "ratelimit",
                        ]
                    ]
                }
            },
        ],
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_ratelimit__over_threshold__minutes(self, _, __, reddit):
        with pytest.raises(RedditAPIException) as exception:
            await reddit.post("test")
        assert (
            exception.value.message
            == "You are doing that too much. Try again in 1 minute."
        )

    @mock.patch(
        "asyncpraw.Reddit.request",
        side_effect=[
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 2 milliseconds.",
                            "ratelimit",
                        ]
                    ]
                }
            },
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 1 millisecond.",
                            "ratelimit",
                        ]
                    ]
                }
            },
            {},
        ],
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_ratelimit__under_threshold__milliseconds(
        self, mock_sleep, _, reddit
    ):
        await reddit.post("test")
        mock_sleep.assert_has_calls([mock.call(1), mock.call(1)])

    @mock.patch(
        "asyncpraw.Reddit.request",
        side_effect=[
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 1 minute.",
                            "ratelimit",
                        ]
                    ]
                }
            },
            {},
        ],
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_ratelimit__under_threshold__minutes(
        self, mock_sleep, _, reddit
    ):
        reddit.config.ratelimit_seconds = 60
        await reddit.post("test")
        mock_sleep.assert_has_calls([mock.call(61)])

    @mock.patch(
        "asyncpraw.Reddit.request",
        side_effect=[
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 5 seconds.",
                            "ratelimit",
                        ]
                    ]
                }
            },
            {},
        ],
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_ratelimit__under_threshold__seconds(
        self, mock_sleep, _, reddit
    ):
        await reddit.post("test")
        mock_sleep.assert_has_calls([mock.call(6)])

    @mock.patch(
        "asyncpraw.Reddit.request",
        side_effect=[
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 5 seconds.",
                            "ratelimit",
                        ]
                    ]
                }
            },
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 3 seconds.",
                            "ratelimit",
                        ]
                    ]
                }
            },
            {
                "json": {
                    "errors": [
                        [
                            "RATELIMIT",
                            "You are doing that too much. Try again in 1 second.",
                            "ratelimit",
                        ]
                    ]
                }
            },
        ],
    )
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_post_ratelimit__under_threshold__seconds_failure(
        self, mock_sleep, _, reddit
    ):
        with pytest.raises(RedditAPIException) as exception:
            await reddit.post("test")
        assert (
            exception.value.message
            == "You are doing that too much. Try again in 1 second."
        )
        mock_sleep.assert_has_calls([mock.call(6), mock.call(4), mock.call(2)])

    async def test_read_only__with_authenticated_core(self):
        async with Reddit(
            token_manager=DummyTokenManager(),
            password=None,
            username=None,
            **self.REQUIRED_DUMMY_SETTINGS,
        ) as reddit:
            assert not reddit.read_only
            reddit.read_only = True
            assert reddit.read_only
            reddit.read_only = False
            assert not reddit.read_only

    def test_read_only__with_authenticated_core__legacy_refresh_token(self):
        with Reddit(
            password=None,
            refresh_token="refresh",
            username=None,
            **self.REQUIRED_DUMMY_SETTINGS,
        ) as reddit:
            assert not reddit.read_only
            reddit.read_only = True
            assert reddit.read_only
            reddit.read_only = False
            assert not reddit.read_only

    async def test_read_only__with_authenticated_core__non_confidential(self):
        async with Reddit(
            token_manager=DummyTokenManager(),
            client_id="dummy",
            client_secret=None,
            redirect_uri="dummy",
            user_agent="dummy",
        ) as reddit:
            assert not reddit.read_only
            reddit.read_only = True
            assert reddit.read_only
            reddit.read_only = False
            assert not reddit.read_only

    def test_read_only__with_authenticated_core__non_confidential__legacy_refresh_token(
        self,
    ):
        with Reddit(
            client_id="dummy",
            client_secret=None,
            redirect_uri="dummy",
            refresh_token="dummy",
            user_agent="dummy",
        ) as reddit:
            assert not reddit.read_only
            reddit.read_only = True
            assert reddit.read_only
            reddit.read_only = False
            assert not reddit.read_only

    async def test_read_only__with_script_authenticated_core(self):
        async with Reddit(
            password="dummy", username="dummy", **self.REQUIRED_DUMMY_SETTINGS
        ) as reddit:
            assert not reddit.read_only
            reddit.read_only = True
            assert reddit.read_only
            reddit.read_only = False
            assert not reddit.read_only

    async def test_read_only__without_trusted_authenticated_core(self):
        async with Reddit(
            password=None, username=None, **self.REQUIRED_DUMMY_SETTINGS
        ) as reddit:
            assert reddit.read_only
            with pytest.raises(ClientException):
                reddit.read_only = False
            assert reddit.read_only
            reddit.read_only = True
            assert reddit.read_only

    async def test_read_only__without_untrusted_authenticated_core(self):
        required_settings = self.REQUIRED_DUMMY_SETTINGS.copy()
        required_settings["client_secret"] = None
        async with Reddit(password=None, username=None, **required_settings) as reddit:
            assert reddit.read_only
            with pytest.raises(ClientException):
                reddit.read_only = False
            assert reddit.read_only
            reddit.read_only = True
            assert reddit.read_only

    def test_reddit__missing_required_settings(self):
        for setting in self.REQUIRED_DUMMY_SETTINGS:
            with pytest.raises(ClientException) as excinfo:
                settings = self.REQUIRED_DUMMY_SETTINGS.copy()
                settings[setting] = Config.CONFIG_NOT_SET
                Reddit(**settings)
            assert str(excinfo.value).startswith(
                f"Required configuration setting {setting!r} missing."
            )
            if setting == "client_secret":
                assert "set to None" in str(excinfo.value)

    def test_reddit__required_settings_set_to_none(self):
        required_settings = self.REQUIRED_DUMMY_SETTINGS.copy()
        del required_settings["client_secret"]
        for setting in required_settings:
            with pytest.raises(ClientException) as excinfo:
                settings = self.REQUIRED_DUMMY_SETTINGS.copy()
                settings[setting] = None
                Reddit(**settings)
            assert str(excinfo.value).startswith(
                f"Required configuration setting {setting!r} missing."
            )

    def test_reddit__site_name_no_section(self):
        with pytest.raises(configparser.NoSectionError) as excinfo:
            Reddit("bad_site_name")
        assert "asyncpraw.readthedocs.io" in excinfo.value.message

    @mock.patch("asyncprawcore.sessions.Session")
    async def test_request__badrequest_with_no_json_body(self, mock_session):
        response = mock.Mock(status=400, text=AsyncMock(return_value=""))
        response.json.side_effect = ValueError
        mock_session.return_value.request = mock.Mock(
            side_effect=BadRequest(response=response)
        )

        reddit = Reddit(client_id="dummy", client_secret="dummy", user_agent="dummy")
        with pytest.raises(Exception) as excinfo:
            await reddit.request(method="POST", path="/")
        assert str(excinfo.value) == "received 400 HTTP response"

    async def test_request__json_and_body(self):
        reddit = Reddit(client_id="dummy", client_secret="dummy", user_agent="dummy")
        with pytest.raises(ClientException) as excinfo:
            await reddit.request(
                data={"key": "value"}, json={"key": "value"}, method="POST", path="/"
            )
        assert str(excinfo.value).startswith(
            "At most one of 'data' or 'json' is supported."
        )

    async def test_submission(self, reddit):
        submission = await reddit.submission("2gmzqe", fetch=False)
        assert submission.id == "2gmzqe"

    async def test_subreddit(self, reddit):
        subreddit = await reddit.subreddit("redditdev")
        assert subreddit.display_name == "redditdev"


class TestRedditCustomRequestor(UnitTest):
    async def test_requestor_class(self, reddit):
        class CustomRequestor(Requestor):
            pass

        async with Reddit(
            requestor_class=CustomRequestor,
            client_id="dummy",
            client_secret="dummy",
            password="dummy",
            user_agent="dummy",
            username="dummy",
        ) as temp_reddit:
            assert isinstance(temp_reddit._core._requestor, CustomRequestor)
        assert not isinstance(reddit._core._requestor, CustomRequestor)

        async with Reddit(
            requestor_class=CustomRequestor,
            client_id="dummy",
            client_secret="dummy",
            user_agent="dummy",
        ) as temp_reddit:
            assert isinstance(temp_reddit._core._requestor, CustomRequestor)
        assert not isinstance(reddit._core._requestor, CustomRequestor)

    async def test_requestor_kwargs(self):
        session = AsyncMock(headers={})
        reddit = Reddit(
            requestor_kwargs={"session": session},
            client_id="dummy",
            client_secret="dummy",
            user_agent="dummy",
        )

        assert reddit._core._requestor._http is session



"""Provide the ReplyableMixin class."""
from typing import TYPE_CHECKING, Optional, Union

from ....const import API_PATH

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class ReplyableMixin:
    """Interface for :class:`.RedditBase` classes that can be replied to."""

    async def reply(
        self, body: str
    ) -> Optional[Union["asyncpraw.models.Comment", "asyncpraw.models.Message"]]:
        """Reply to the object.

        :param body: The Markdown formatted content for a comment.

        :returns: A :class:`.Comment` or :class:`.Message` object for the newly created
            comment or message or ``None`` if Reddit doesn't provide one.

        :raises: ``asyncprawcore.exceptions.Forbidden`` when attempting to reply to some
            items, such as locked submissions/comments or non-replyable messages.

        A ``None`` value can be returned if the target is a comment or submission in a
        quarantined subreddit and the authenticated user has not opt-ed in to viewing
        the content. When this happens the comment will be successfully created on
        Reddit and can be retried by drawing the comment from the user's comment
        history.

        Example usage:

        .. code-block:: python

            submission = await reddit.submission("5or86n", fetch=False)
            await submission.reply("reply")

            comment = await reddit.comment("dxolpyc", fetch=False)
            await comment.reply("reply")

        """
        data = {"text": body, "thing_id": self.fullname}
        comments = await self._reddit.post(API_PATH["comment"], data=data)
        try:
            return comments[0]
        except IndexError:  # pragma: no cover; I haven't been able to make this happen again
            return None



import pytest

from asyncpraw.models import Draft, Subreddit

from ... import UnitTest


class TestDraft(UnitTest):
    def test_construct_failure(self, reddit):
        message = "Exactly one of 'id' or '_data' must be provided."
        with pytest.raises(TypeError) as excinfo:
            Draft(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Draft(reddit, "dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Draft(reddit, id="dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

    async def test_create_failure(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            await reddit.drafts.create(url="url", selftext="selftext")
        assert (
            str(excinfo.value) == "Exactly one of 'selftext' or 'url' must be provided."
        )

    def test_equality(self, reddit):
        draft1 = Draft(reddit, _data={"id": "dummy1"})
        draft2 = Draft(reddit, _data={"id": "dummy1"})
        draft3 = Draft(reddit, _data={"id": "dummy3"})
        assert draft1 == draft1
        assert draft2 == draft2
        assert draft3 == draft3
        assert draft1 == draft2
        assert draft2 != draft3
        assert draft1 != draft3
        assert "dummy1" == draft1
        assert draft2 == "dummy1"

        draft1 = Draft(
            reddit, _data={"id": "dummy1", "body": "body1", "kind": "markdown"}
        )
        draft2 = Draft(
            reddit, _data={"id": "dummy1", "body": "body1", "kind": "markdown"}
        )
        draft3 = Draft(
            reddit, _data={"id": "dummy3", "body": "body2", "kind": "markdown"}
        )
        assert draft1 == draft1
        assert draft2 == draft2
        assert draft3 == draft3
        assert draft1 == draft2
        assert draft2 != draft3
        assert draft1 != draft3

        draft1 = Draft(reddit, _data={"id": "dummy1", "body": "url1", "kind": "link"})
        draft2 = Draft(reddit, _data={"id": "dummy1", "body": "url1", "kind": "link"})
        draft3 = Draft(reddit, _data={"id": "dummy3", "body": "url3", "kind": "link"})
        assert draft1 == draft1
        assert draft2 == draft2
        assert draft3 == draft3
        assert draft1 == draft2
        assert draft2 != draft3
        assert draft1 != draft3

    def test_hash(self, reddit):
        draft1 = Draft(
            reddit, _data={"id": "dummy1", "body": "body1", "kind": "markdown"}
        )
        draft2 = Draft(
            reddit, _data={"id": "dummy1", "body": "body2", "kind": "markdown"}
        )
        draft3 = Draft(
            reddit, _data={"id": "dummy3", "body": "body2", "kind": "markdown"}
        )
        assert hash(draft1) == hash(draft1)
        assert hash(draft2) == hash(draft2)
        assert hash(draft3) == hash(draft3)
        assert hash(draft1) == hash(draft2)
        assert hash(draft2) != hash(draft3)
        assert hash(draft1) != hash(draft3)

    def test_repr(self, reddit):
        draft = Draft(reddit, id="draft_id")
        assert repr(draft) == "Draft(id='draft_id')"

        data = {"id": "draft_id", "body": "body", "kind": "markdown"}
        subreddit = Subreddit(None, "subreddit")
        draft = Draft(reddit, _data={**data, "subreddit": subreddit, "title": None})
        assert repr(draft) == "Draft(id='draft_id' subreddit='subreddit')"

        draft = Draft(reddit, _data={**data, "subreddit": None, "title": "title"})
        assert repr(draft) == "Draft(id='draft_id' title='title')"

        draft = Draft(reddit, _data={**data, "subreddit": subreddit, "title": "title"})
        assert repr(draft) == "Draft(id='draft_id' subreddit='subreddit' title='title')"

    def test_str(self, reddit):
        draft = Draft(reddit, _data={"id": "dummy"})
        assert str(draft) == "dummy"

    async def test_submit_failure(self, reddit):
        draft = Draft(
            reddit,
            _data={
                "id": "draft_id",
                "body": "body",
                "kind": "markdown",
                "subreddit": None,
            },
        )
        with pytest.raises(ValueError) as excinfo:
            await draft.submit()
            assert (
                str(excinfo.value)
                == "'subreddit' must be set on the Draft or passed as a keyword argument."
            )



"""Token Manager classes.

There should be a 1-to-1 mapping between an instance of a subclass of
:class:`.BaseTokenManager` and a :class:`.Reddit` instance.

A few proof of concept token manager classes are provided here, but it is expected that
Async PRAW users will create their own token manager classes suitable for their needs.

.. deprecated:: 7.4.0

    Tokens managers have been deprecated and will be removed in the near future.

"""
from abc import ABC, abstractmethod

import aiofiles
from asyncio_extras import async_contextmanager

from . import _deprecate_args


class BaseTokenManager(ABC):
    """An abstract class for all token managers."""

    def __init__(self):
        """Initialize a :class:`.BaseTokenManager` instance."""
        self._reddit = None

    @property
    def reddit(self):
        """Return the :class:`.Reddit` instance bound to the token manager."""
        return self._reddit

    @reddit.setter
    def reddit(self, value):
        if self._reddit is not None:
            raise RuntimeError(
                "'reddit' can only be set once and is done automatically"
            )
        self._reddit = value

    @abstractmethod
    def post_refresh_callback(self, authorizer):
        """Handle callback that is invoked after a refresh token is used.

        :param authorizer: The ``asyncprawcore.Authorizer`` instance used containing
            ``access_token`` and ``refresh_token`` attributes.

        This function will be called after refreshing the access and refresh tokens.
        This callback can be used for saving the updated ``refresh_token``.

        """

    @abstractmethod
    def pre_refresh_callback(self, authorizer):
        """Handle callback that is invoked before refreshing PRAW's authorization.

        :param authorizer: The ``asyncprawcore.Authorizer`` instance used containing
            ``access_token`` and ``refresh_token`` attributes.

        This callback can be used to inspect and modify the attributes of the
        ``asyncprawcore.Authorizer`` instance, such as setting the ``refresh_token``.

        """


class FileTokenManager(BaseTokenManager):
    """Provides a single-file based token manager.

    It is expected that the file with the initial ``refresh_token`` is created prior to
    use.

    .. warning::

        The same ``file`` should not be used by more than one instance of this class
        concurrently. Doing so may result in data corruption. Consider using
        :class:`.SQLiteTokenManager` if you want more than one instance of PRAW to
        concurrently manage a specific ``refresh_token`` chain.

    """

    def __init__(self, filename):
        """Initialize a :class:`.FileTokenManager` instance.

        :param filename: The file the contains the refresh token.

        """
        super().__init__()
        self._filename = filename

    async def post_refresh_callback(self, authorizer):
        """Update the saved copy of the refresh token."""
        async with aiofiles.open(self._filename, "w") as fp:
            await fp.write(authorizer.refresh_token)

    async def pre_refresh_callback(self, authorizer):
        """Load the refresh token from the file."""
        if authorizer.refresh_token is None:
            async with aiofiles.open(self._filename) as fp:
                authorizer.refresh_token = (await fp.read()).strip()


class SQLiteTokenManager(BaseTokenManager):
    """Provides a SQLite3 based token manager.

    Unlike, :class:`.FileTokenManager`, the initial database need not be created ahead
    of time, as it'll automatically be created on first use. However, initial refresh
    tokens will need to be registered via :meth:`.register` prior to use.

    .. warning::

        This class is untested on Windows because we encountered file locking issues in
        the test environment.

    """

    @_deprecate_args("database", "key")
    def __init__(self, *, database, key):
        """Initialize a :class:`.SQLiteTokenManager` instance.

        :param database: The path to the SQLite database.
        :param key: The key used to locate the refresh token. This ``key`` can be
            anything. You might use the ``client_id`` if you expect to have unique a
            refresh token for each ``client_id``, or you might use a redditor's
            ``username`` if you're managing multiple users' authentications.

        """
        super().__init__()
        self._connection = None
        self._database = database
        self._setup_ran = False
        self.key = key

    @async_contextmanager
    async def connection(self):
        """Asynchronously setup and provide the sqlite3 connection."""
        if self._connection is None:
            import aiosqlite

            self._connection = await aiosqlite.connect(self._database)
        if not self._setup_ran:
            await self._connection.execute(
                "CREATE TABLE IF NOT EXISTS tokens (id, refresh_token, updated_at)"
            )
            await self._connection.execute(
                "CREATE UNIQUE INDEX IF NOT EXISTS ux_tokens_id on tokens(id)"
            )
            await self._connection.commit()
            self._setup_ran = True
        yield self._connection

    async def _get(self):
        async with self.connection() as conn:
            cursor = await conn.execute(
                "SELECT refresh_token FROM tokens WHERE id=?", (self.key,)
            )
            result = await cursor.fetchone()
            if result is None:
                raise KeyError
        return result[0]

    async def _set(self, refresh_token):
        """Set the refresh token in the database.

        This function will overwrite an existing value if the corresponding ``key``
        already exists.

        """
        async with self.connection() as conn:
            await conn.execute(
                "REPLACE INTO tokens VALUES (?, ?, datetime('now'))",
                (self.key, refresh_token),
            )
            await conn.commit()

    async def close(self):
        """Close the sqlite3 connection."""
        await self._connection.close()

    async def is_registered(self):
        """Return whether ``key`` already has a ``refresh_token``."""
        async with self.connection() as conn:
            cursor = await conn.execute(
                "SELECT refresh_token FROM tokens WHERE id=?", (self.key,)
            )
            result = await cursor.fetchone()
        return result is not None

    async def post_refresh_callback(self, authorizer):
        """Update the refresh token in the database."""
        await self._set(authorizer.refresh_token)

        # While the following line is not strictly necessary, it ensures that the
        # refresh token is not used elsewhere. And also forces the pre_refresh_callback
        # to always load the latest refresh_token from the database.
        authorizer.refresh_token = None

    async def pre_refresh_callback(self, authorizer):
        """Load the refresh token from the database."""
        assert authorizer.refresh_token is None
        authorizer.refresh_token = await self._get()

    async def register(self, refresh_token):
        """Register the initial refresh token in the database.

        :returns: ``True`` if ``refresh_token`` is saved to the database, otherwise,
            ``False`` if there is already a ``refresh_token`` for the associated
            ``key``.

        """
        async with self.connection() as conn:
            cursor = await conn.execute(
                "INSERT OR IGNORE INTO tokens VALUES (?, ?, datetime('now'))",
                (self.key, refresh_token),
            )
            await conn.commit()
            row_count = cursor.rowcount
        return row_count == 1



"""Prepare py.test."""
import asyncio
import json
import os
from base64 import b64encode
from datetime import datetime
from functools import wraps

import pytest
from _pytest.tmpdir import _mk_tmp
from vcr import VCR
from vcr.cassette import Cassette
from vcr.persisters.filesystem import FilesystemPersister
from vcr.serialize import deserialize, serialize


# Prevent calls to sleep
async def _sleep(*args):
    raise Exception("Call to sleep")


asyncio.sleep = _sleep


def b64_string(input_string):
    """Return a base64 encoded string (not bytes) from input_string."""
    return b64encode(input_string.encode("utf-8")).decode("utf-8")


def env_default(key):
    """Return environment variable or placeholder string."""
    return os.environ.get(f"prawtest_{key}", f"placeholder_{key}")


def filter_access_token(response):
    """Add VCR callback to filter access token."""
    request_uri = response["url"]
    if "api/v1/access_token" not in request_uri or response["status"]["code"] != 200:
        return response
    body = response["body"]["string"].decode()
    try:
        token = json.loads(body)["access_token"]
        response["body"]["string"] = response["body"]["string"].replace(
            token.encode("utf-8"), b"<ACCESS_TOKEN>"
        )
        placeholders["access_token"] = token
    except (KeyError, TypeError, ValueError):
        pass
    return response


def serialize_dict(data: dict):
    """This is to filter out buffered readers."""
    new_dict = {}
    for key, value in data.items():
        if key == "file":
            new_dict[key] = serialize_file(value.name)
        elif isinstance(value, dict):
            new_dict[key] = serialize_dict(value)
        elif isinstance(value, list):
            new_dict[key] = serialize_list(value)
        else:
            new_dict[key] = value
    return new_dict


def serialize_file(file_name):
    with open(file_name, "rb") as f:
        return f.read().decode("utf-8", "replace")


def serialize_list(data: list):
    new_list = []
    for item in data:
        if isinstance(item, dict):
            new_list.append(serialize_dict(item))
        elif isinstance(item, list):
            new_list.append(serialize_list(item))
        elif isinstance(item, tuple):
            if item[0] == "file":
                item = (item[0], serialize_file(item[1].name))
            new_list.append(item)
        else:
            new_list.append(item)
    return new_list


placeholders = {
    x: env_default(x)
    for x in (
        "auth_code client_id client_secret password redirect_uri test_subreddit"
        " user_agent username refresh_token"
    ).split()
}

placeholders["basic_auth"] = b64_string(
    f"{placeholders['client_id']}:{placeholders['client_secret']}"
)


class CustomPersister(FilesystemPersister):
    @classmethod
    def load_cassette(cls, cassette_path, serializer):
        try:
            with open(cassette_path) as f:
                cassette_content = f.read()
        except OSError:
            raise ValueError("Cassette not found.")
        for replacement, value in [
            (v, f"<{k.upper()}>") for k, v in placeholders.items()
        ]:
            cassette_content = cassette_content.replace(value, replacement)
        cassette = deserialize(cassette_content, serializer)
        return cassette

    @staticmethod
    def save_cassette(cassette_path, cassette_dict, serializer):
        data = serialize(cassette_dict, serializer)
        for replacement, value in [
            (f"<{k.upper()}>", v) for k, v in placeholders.items()
        ]:
            data = data.replace(value, replacement)
        dirname, filename = os.path.split(cassette_path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(cassette_path, "w") as f:
            f.write(data)


class CustomSerializer(object):
    @staticmethod
    def serialize(cassette_dict):
        cassette_dict["recorded_at"] = datetime.now().isoformat()[:-7]
        return (
            f"{json.dumps(serialize_dict(cassette_dict), sort_keys=True, indent=2)}\n"
        )

    @staticmethod
    def deserialize(cassette_string):
        return json.loads(cassette_string)


vcr = VCR(
    before_record_response=filter_access_token,
    cassette_library_dir="tests/integration/cassettes",
    match_on=["uri", "method"],
    path_transformer=VCR.ensure_suffix(".json"),
    serializer="custom_serializer",
)
vcr.register_serializer("custom_serializer", CustomSerializer)
vcr.register_persister(CustomPersister)


def after_init(func, *args):
    func(*args)


def add_init_hook(original_init):
    """Wrap an __init__ method to also call some hooks."""

    @wraps(original_init)
    def wrapper(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        after_init(init_hook, self)

    return wrapper


Cassette.__init__ = add_init_hook(Cassette.__init__)


def init_hook(cassette):
    if not cassette.requests:
        pytest.set_up_record()  # dynamically defined in __init__.py


class Placeholders:
    def __init__(self, _dict):
        self.__dict__ = _dict


def pytest_configure():
    pytest.placeholders = Placeholders(placeholders)


@pytest.fixture
def tmp_path(request, tmp_path_factory):
    # Manually create tmp_path fixture since asynctest does not play nicely with
    # fixtures as args
    request.cls.tmp_path = _mk_tmp(request, tmp_path_factory)



"""Provide the ModeratedList class."""
from .base import BaseList


class ModeratedList(BaseList):
    """A list of moderated :class:`.Subreddit` objects. Works just like a regular list."""

    CHILD_ATTRIBUTE = "data"



"""Provide the Redditor class."""
from json import dumps
from typing import TYPE_CHECKING, Any, AsyncGenerator, Dict, List, Optional, Union

from ...const import API_PATH
from ...util import _deprecate_args
from ...util.cache import cachedproperty
from ..listing.mixins import RedditorListingMixin
from ..util import stream_generator
from .base import RedditBase
from .mixins import FullnameMixin, MessageableMixin

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class Redditor(MessageableMixin, RedditorListingMixin, FullnameMixin, RedditBase):
    """A class representing the users of Reddit.

    .. include:: ../../typical_attributes.rst

    .. note::

        Shadowbanned accounts are treated the same as non-existent accounts, meaning
        that they will not have any attributes.

    .. note::

        Suspended/banned accounts will only return the ``name`` and ``is_suspended``
        attributes.

    =================================== ================================================
    Attribute                           Description
    =================================== ================================================
    ``comment_karma``                   The comment karma for the :class:`.Redditor`.
    ``comments``                        Provide an instance of :class:`.SubListing` for
                                        comment access.
    ``submissions``                     Provide an instance of :class:`.SubListing` for
                                        submission access.
    ``created_utc``                     Time the account was created, represented in
                                        `Unix Time`_.
    ``has_verified_email``              Whether or not the :class:`.Redditor` has
                                        verified their email.
    ``icon_img``                        The url of the Redditors' avatar.
    ``id``                              The ID of the :class:`.Redditor`.
    ``is_employee``                     Whether or not the :class:`.Redditor` is a
                                        Reddit employee.
    ``is_friend``                       Whether or not the :class:`.Redditor` is friends
                                        with the authenticated user.
    ``is_mod``                          Whether or not the :class:`.Redditor` mods any
                                        subreddits.
    ``is_gold``                         Whether or not the :class:`.Redditor` has active
                                        Reddit Premium status.
    ``is_suspended``                    Whether or not the :class:`.Redditor` is
                                        currently suspended.
    ``link_karma``                      The link karma for the :class:`.Redditor`.
    ``name``                            The Redditor's username.
    ``subreddit``                       If the :class:`.Redditor` has created a
                                        user-subreddit, provides a dictionary of
                                        additional attributes. See below.
    ``subreddit["banner_img"]``         The URL of the user-subreddit banner.
    ``subreddit["name"]``               The fullname of the user-subreddit.
    ``subreddit["over_18"]``            Whether or not the user-subreddit is NSFW.
    ``subreddit["public_description"]`` The public description of the user- subreddit.
    ``subreddit["subscribers"]``        The number of users subscribed to the
                                        user-subreddit.
    ``subreddit["title"]``              The title of the user-subreddit.
    =================================== ================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    STR_FIELD = "name"

    @classmethod
    def from_data(cls, reddit, data):
        """Return an instance of :class:`.Redditor`, or ``None`` from ``data``."""
        if data == "[deleted]":
            return None
        return cls(reddit, data)

    @cachedproperty
    def notes(self) -> "asyncpraw.models.RedditorModNotes":
        """Provide an instance of :class:`.RedditorModNotes`.

        This provides an interface for managing moderator notes for a redditor.

        .. note::

            The authenticated user must be a moderator of the provided subreddit(s).

        For example, all the notes for u/spez in r/test can be iterated through like so:

        .. code-block:: python

            redditor = await reddit.redditor("spez")

            async for note in redditor.notes.subreddits("test"):
                print(f"{note.label}: {note.note}")

        """
        from ..mod_notes import RedditorModNotes

        return RedditorModNotes(self._reddit, self)

    @cachedproperty
    def stream(self) -> "asyncpraw.models.reddit.redditor.RedditorStream":
        """Provide an instance of :class:`.RedditorStream`.

        Streams can be used to indefinitely retrieve new comments made by a redditor,
        like:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for comment in redditor.stream.comments():
                print(comment)

        Additionally, new submissions can be retrieved via the stream. In the following
        example all submissions are fetched via the redditor u/spez:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for submission in redditor.stream.submissions():
                print(submission)

        """
        return RedditorStream(self)

    @property
    def _kind(self) -> str:
        """Return the class's kind."""
        return self._reddit.config.kinds["redditor"]

    @property
    def _path(self) -> str:
        return API_PATH["user"].format(user=self)

    def __init__(
        self,
        reddit: "asyncpraw.Reddit",
        name: Optional[str] = None,
        fullname: Optional[str] = None,
        _data: Optional[Dict[str, Any]] = None,
    ):
        """Initialize a :class:`.Redditor` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param name: The name of the redditor.
        :param fullname: The fullname of the redditor, starting with ``t2_``.

        Exactly one of ``name``, ``fullname`` or ``_data`` must be provided.

        """
        if (name, fullname, _data).count(None) != 2:
            raise TypeError(
                "Exactly one of 'name', 'fullname', or '_data' must be provided."
            )
        if _data:
            assert (
                isinstance(_data, dict) and "name" in _data
            ), "Please file a bug with Async PRAW."
        self._listing_use_sort = True
        if name:
            self.name = name
        elif fullname:
            self._fullname = fullname
        super().__init__(reddit, _data=_data, _extra_attribute_to_check="_fullname")

    def __setattr__(self, name: str, value: Any):
        """Objectify the subreddit attribute."""
        if name == "subreddit" and value:
            from .user_subreddit import UserSubreddit

            value = UserSubreddit(reddit=self._reddit, _data=value)
        super().__setattr__(name, value)

    async def _fetch_username(self, fullname):
        response = await self._reddit.get(
            API_PATH["user_by_fullname"], params={"ids": fullname}
        )
        return response[fullname]["name"]

    async def _fetch_info(self):
        if hasattr(self, "_fullname"):
            self.name = await self._fetch_username(self._fullname)
        return "user_about", {"user": self.name}, None

    async def _fetch_data(self):
        name, fields, params = await self._fetch_info()
        path = API_PATH[name].format(**fields)
        return await self._reddit.request(method="GET", params=params, path=path)

    async def _fetch(self):
        data = await self._fetch_data()
        data = data["data"]
        other = type(self)(self._reddit, _data=data)
        self.__dict__.update(other.__dict__)
        self._fetched = True

    async def _friend(self, *, data, method):
        url = API_PATH["friend_v1"].format(user=self)
        await self._reddit.request(data=dumps(data), method=method, path=url)

    async def block(self):
        """Block the :class:`.Redditor`.

        For example, to block :class:`.Redditor` u/spez:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            await redditor.block()

        .. note::

            Blocking a trusted user will remove that user from your trusted list.

        .. seealso::

            :meth:`.trust`

        """
        await self._reddit.post(API_PATH["block_user"], params={"name": self.name})

    async def distrust(self):
        """Remove the :class:`.Redditor` from your whitelist of trusted users.

        For example, to remove :class:`.Redditor` u/spez from your whitelist:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            await redditor.distrust()

        .. seealso::

            :meth:`.trust`

        """
        await self._reddit.post(
            API_PATH["remove_whitelisted"], data={"name": self.name}
        )

    @_deprecate_args("note")
    async def friend(self, *, note: str = None):
        """Friend the :class:`.Redditor`.

        :param note: A note to save along with the relationship. Requires Reddit Premium
            (default: ``None``).

        Calling this method subsequent times will update the note.

        For example, to friend u/spez:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            await redditor.friend()

        To add a note to the friendship (requires Reddit Premium):

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            await redditor.friend(note="My favorite admin")

        """
        await self._friend(data={"note": note} if note else {}, method="PUT")

    async def friend_info(self) -> "asyncpraw.models.Redditor":
        """Return a :class:`.Redditor` instance with specific friend-related attributes.

        :returns: A :class:`.Redditor` instance with fields ``date``, ``id``, and
            possibly ``note`` if the authenticated user has Reddit Premium.

        For example, to get the friendship information of :class:`.Redditor` u/spez:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            info = await redditor.friend_info
            friend_data = info.date

        """
        return await self._reddit.get(API_PATH["friend_v1"].format(user=self))

    @_deprecate_args("months")
    async def gild(self, *, months: int = 1):
        """Gild the :class:`.Redditor`.

        :param months: Specifies the number of months to gild up to 36 (default: ``1``).

        For example, to gild :class:`.Redditor` u/spez for 1 month:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            await redditor.gild(months=1)

        """
        if months < 1 or months > 36:
            raise TypeError("months must be between 1 and 36")
        await self._reddit.post(
            API_PATH["gild_user"].format(username=self), data={"months": months}
        )

    async def moderated(self) -> List["asyncpraw.models.Subreddit"]:
        """Return a list of the redditor's moderated subreddits.

        :returns: A list of :class:`.Subreddit` objects. Return ``[]`` if the redditor
            has no moderated subreddits.

        :raises: ``asyncprawcore.ServerError`` in certain circumstances. See the note
            below.

        .. note::

            The redditor's own user profile subreddit will not be returned, but other
            user profile subreddits they moderate will be returned.

        Usage:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for subreddit in redditor.moderated():
                print(subreddit.display_name)
                print(subreddit.title)

        .. note::

            A ``asyncprawcore.ServerError`` exception may be raised if the redditor
            moderates a large number of subreddits. If that happens, try switching to
            :ref:`read-only mode <read_only_application>`. For example,

            .. code-block:: python

                reddit.read_only = True
                redditor = await reddit.redditor("reddit")
                async for subreddit in redditor.moderated():
                    print(str(subreddit))

            It is possible that requests made in read-only mode will also raise a
            ``asyncprawcore.ServerError`` exception.

            When used in read-only mode, this method does not retrieve information about
            subreddits that require certain special permissions to access, e.g., private
            subreddits and premium-only subreddits.

        .. seealso::

            :meth:`.User.moderator_subreddits`

        """
        return await self._reddit.get(API_PATH["moderated"].format(user=self)) or []

    async def multireddits(self) -> List["asyncpraw.models.Multireddit"]:
        """Return a list of the redditor's public multireddits.

        For example, to to get :class:`.Redditor` u/spez's multireddits:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            multireddits = await redditor.multireddits()

        """
        return await self._reddit.get(API_PATH["multireddit_user"].format(user=self))

    async def trophies(self) -> List["asyncpraw.models.Trophy"]:
        """Return a list of the redditor's trophies.

        :returns: A list of :class:`.Trophy` objects. Return ``[]`` if the redditor has
            no trophies.

        :raises: :class:`.RedditAPIException` if the redditor doesn't exist.

        Usage:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for trophy in redditor.trophies():
                print(trophy.name)
                print(trophy.description)

        """
        return list(await self._reddit.get(API_PATH["trophies"].format(user=self)))

    async def trust(self):
        """Add the :class:`.Redditor` to your whitelist of trusted users.

        Trusted users will always be able to send you PMs.

        Example usage:

        .. code-block:: python

            redditor = await reddit.redditor("AaronSw")
            await redditor.trust()

        Use the ``accept_pms`` parameter of :meth:`.Preferences.update` to toggle your
        ``accept_pms`` setting between ``"everyone"`` and ``"whitelisted"``. For
        example:

        .. code-block:: python

            # Accept private messages from everyone:
            await reddit.user.preferences.update(accept_pms="everyone")
            # Only accept private messages from trusted users:
            await reddit.user.preferences.update(accept_pms="whitelisted")

        You may trust a user even if your ``accept_pms`` setting is switched to
        ``"everyone"``.

        .. note::

            You are allowed to have a user on your blocked list and your friends list at
            the same time. However, you cannot trust a user who is on your blocked list.

        .. seealso::

            - :meth:`.distrust`
            - :meth:`.Preferences.update`
            - :meth:`.trusted`

        """
        await self._reddit.post(API_PATH["add_whitelisted"], data={"name": self.name})

    async def unblock(self):
        """Unblock the :class:`.Redditor`.

        For example, to unblock :class:`.Redditor` u/spez:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            await redditor.unblock()

        """
        container = await self._reddit.user.me()
        data = {
            "container": container.fullname,
            "name": str(self),
            "type": "enemy",
        }
        url = API_PATH["unfriend"].format(subreddit="all")
        await self._reddit.post(url, data=data)

    async def unfriend(self):
        """Unfriend the :class:`.Redditor`.

        For example, to unfriend :class:`.Redditor` u/spez:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            await redditor.unfriend()

        """
        await self._friend(data={"id": str(self)}, method="DELETE")


class RedditorStream:
    """Provides submission and comment streams."""

    def __init__(self, redditor: "asyncpraw.models.Redditor"):
        """Initialize a :class:`.RedditorStream` instance.

        :param redditor: The redditor associated with the streams.

        """
        self.redditor = redditor

    def comments(
        self, **stream_options: Union[str, int, Dict[str, str]]
    ) -> AsyncGenerator["asyncpraw.models.Comment", None]:
        """Yield new comments as they become available.

        Comments are yielded oldest first. Up to 100 historical comments will initially
        be returned.

        Keyword arguments are passed to :func:`.stream_generator`.

        For example, to retrieve all new comments made by redditor u/spez, try:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for comment in redditor.stream.comments():
                print(comment)

        """
        return stream_generator(self.redditor.comments.new, **stream_options)

    def submissions(
        self, **stream_options: Union[str, int, Dict[str, str]]
    ) -> AsyncGenerator["asyncpraw.models.Submission", None]:
        """Yield new submissions as they become available.

        Submissions are yielded oldest first. Up to 100 historical submissions will
        initially be returned.

        Keyword arguments are passed to :func:`.stream_generator`.

        For example, to retrieve all new submissions made by redditor u/spez, try:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for submission in redditor.stream.submissions():
                print(submission)

        """
        return stream_generator(self.redditor.submissions.new, **stream_options)



from unittest import mock

import pytest

from asyncpraw.exceptions import RedditAPIException
from tests.integration import IntegrationTest


class TestUserSubredditModeration(IntegrationTest):
    async def test_accept_invite__no_invite(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                user = await self.reddit.redditor("Lil_SpazJoekp", fetch=True)
                await user.subreddit.mod.accept_invite()
            assert excinfo.value.items[0].error_type == "NO_INVITE_FOUND"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            authenticated_user = await self.reddit.user.me()
            before_settings = await authenticated_user.subreddit.mod.settings()
            new_title = f"{before_settings['title']}x"
            new_title = (
                "x"
                if (len(new_title) >= 20 and "placeholder" not in new_title)
                else new_title
            )
            await authenticated_user.subreddit.mod.update(title=new_title)
            authenticated_user = await self.reddit.user.me(use_cache=False)
            assert authenticated_user.subreddit.title == new_title
            after_settings = await authenticated_user.subreddit.mod.settings()

            # Ensure that nothing has changed besides what was specified.
            before_settings["title"] = new_title
            assert before_settings == after_settings



"""Asynchronous Python Reddit API Wrapper.

Async PRAW, an abbreviation for "Asynchronous Python Reddit API Wrapper", is a python
package that allows for simple access to Reddit's API. Async PRAW aims to be as easy to
use as possible and is designed to follow all of Reddit's API rules. You have to give an
useragent, everything else is handled by Async PRAW so you needn't worry about violating
them.

More information about Async PRAW can be found at https://github.com/praw-dev/asyncpraw

"""

from .const import __version__  # NOQA
from .reddit import Reddit  # NOQA



"""Test asyncpraw.models.user."""
import sys

import asyncprawcore.exceptions
import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import RedditAPIException
from asyncpraw.models import Multireddit, Redditor, Submission, Subreddit

from .. import IntegrationTest


class TestUser(IntegrationTest):
    async def test_blocked(self):
        self.reddit.read_only = False
        with self.use_cassette():
            blocked = await self.reddit.user.blocked()
        assert len(blocked) > 0
        assert all(isinstance(user, Redditor) for user in blocked)

    async def test_blocked_fullname(self):
        self.reddit.read_only = False
        with self.use_cassette():
            blocked = next(iter(await self.reddit.user.blocked()))
            assert blocked.fullname.startswith("t2_")
            assert not blocked.fullname.startswith("t2_t2")

    async def test_contributor_subreddits(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for subreddit in self.reddit.user.contributor_subreddits():
                assert isinstance(subreddit, Subreddit)
                count += 1
            assert count > 0

    async def test_friends(self):
        self.reddit.read_only = False
        with self.use_cassette():
            friends = await self.reddit.user.friends()
        assert len(friends) > 0
        assert all(isinstance(friend, Redditor) for friend in friends)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_friend_exist(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            friend = await self.reddit.user.friends(user=await self.reddit.user.me())
            assert isinstance(friend, Redditor)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_friend_not_exist(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException):
                await self.reddit.user.friends(user="fake__user_user_user")

    async def test_karma(self):
        self.reddit.read_only = False
        with self.use_cassette():
            karma = await self.reddit.user.karma()
        assert isinstance(karma, dict)
        for subreddit in karma:
            assert isinstance(subreddit, Subreddit)
            keys = sorted(karma[subreddit].keys())
            assert ["comment_karma", "link_karma"] == keys

    async def test_me(self):
        self.reddit.read_only = False
        with self.use_cassette():
            me = await self.reddit.user.me()
        assert isinstance(me, Redditor)
        me.praw_is_cached = True
        me = await self.reddit.user.me()
        assert me.praw_is_cached

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_me__bypass_cache(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            me = await self.reddit.user.me()
            me.praw_is_cached = True
            me = await self.reddit.user.me(use_cache=False)
            assert not hasattr(me, "Async PRAW_is_cached")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_moderator_subreddits(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            mod_subs = await self.async_list(
                self.reddit.user.moderator_subreddits(limit=None)
            )
            assert mod_subs
            assert all(isinstance(x, Subreddit) for x in mod_subs)

    async def test_multireddits(self):
        self.reddit.read_only = False
        with self.use_cassette():
            multireddits = await self.reddit.user.multireddits()
            assert isinstance(multireddits, list)
            assert multireddits
            assert all(isinstance(x, Multireddit) for x in multireddits)

    async def test_pin(self):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission_list = []
            for i in range(1, 5):
                submission = await subreddit.submit(
                    title=f"Async PRAW Test {i}", selftext=f"Testing .pin method {i}"
                )
                submission_list.append(submission)
                await self.reddit.user.pin(submission)

            for i in range(5, 9):
                await subreddit.submit(
                    title=f"Async PRAW Test {i}", selftext=f"Testing .pin method {i}"
                )
            new_posts = await self.async_list(
                (await self.reddit.user.me()).new(limit=4)
            )
            new_posts.reverse()
            assert new_posts == submission_list

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_pin__comment(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            comment = await self.reddit.comment("hnxx8f2")
            await self.reddit.user.pin(comment)
            new_content = await self.async_next(
                (await self.reddit.user.me()).new(limit=1)
            )
            assert new_content != comment

    async def test_pin__deleted_submission(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(asyncprawcore.exceptions.BadRequest):
                await self.reddit.user.pin(Submission(self.reddit, "rmhl6m"))

    async def test_pin__empty_slot(self):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            new_posts = await self.async_list(
                (await self.reddit.user.me()).new(limit=4)
            )
            new_posts.reverse()
            for i in range(2, 4):
                await self.reddit.user.pin(new_posts[i], state=False)
            submission = await subreddit.submit(
                title="Async PRAW Test 5", selftext="Testing .pin method 5"
            )
            await self.reddit.user.pin(submission, num=4)
            new_posts = await self.async_list(
                (await self.reddit.user.me()).new(limit=4)
            )
            new_posts.reverse()
            assert new_posts[-1] == submission

    async def test_pin__ignore_conflicts(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await self.reddit.user.pin(Submission(self.reddit, "rmi79w"))
            await self.reddit.user.pin(Submission(self.reddit, "rmi79w"))

    async def test_pin__invalid_num(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await self.reddit.user.pin(Submission(self.reddit, "rmi7hx"), num=7)
            submission = await self.async_next(
                (await self.reddit.user.me()).new(limit=1)
            )
            assert submission.id == "rmi7hx"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_pin__num(self, _):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission_list = []
            for i in range(1, 5):
                submission = await subreddit.submit(
                    title=f"Async PRAW Test {i}", selftext=f"Testing .pin method {i}"
                )
                submission_list.append(submission)
            submission_list.reverse()
            for num, submission in enumerate(submission_list, 1):
                await self.reddit.user.pin(submission, num=num)

            new_posts = await self.async_list(
                (await self.reddit.user.me()).new(limit=4)
            )
            assert new_posts == submission_list

    async def test_pin__remove(self):
        self.reddit.read_only = False
        with self.use_cassette():
            unpinned_posts = set()
            async for post in (await self.reddit.user.me()).new(limit=4):
                await self.reddit.user.pin(post, state=False)
                unpinned_posts.add(post.title)
            new_posts = set(
                [
                    submission.title
                    async for submission in (await self.reddit.user.me()).new(limit=4)
                ]
            )
            assert unpinned_posts != new_posts

    async def test_pin__remove_num(self):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            await self.reddit.user.pin(
                Submission(self.reddit, "rmi7ab"), num=1, state=False
            )
            submission = await self.async_next(
                (await self.reddit.user.me()).new(limit=1)
            )
            assert submission.id != "rmi7ab"

    async def test_pin__removed_submission(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(asyncprawcore.exceptions.BadRequest):
                await self.reddit.user.pin(Submission(self.reddit, "rmi7ab"))

    async def test_pin__replace_slot(self):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            submission = await subreddit.submit(
                title="Async PRAW Test replace slot 1", selftext="Testing .pin method 1"
            )
            await self.reddit.user.pin(submission, num=1)
            new_posts = await self.async_list(
                (await self.reddit.user.me()).new(limit=4)
            )
            new_posts.reverse()
            assert new_posts[-1] == submission

    async def test_subreddits(self):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            async for subreddit in self.reddit.user.subreddits():
                assert isinstance(subreddit, Subreddit)
                count += 1
            assert count > 0



import pytest

from asyncpraw.exceptions import ClientException
from asyncpraw.models import Submission

from ... import UnitTest


class TestSubmission(UnitTest):
    def test_equality(self, reddit):
        submission1 = Submission(reddit, _data={"id": "dummy1", "n": 1})
        submission2 = Submission(reddit, _data={"id": "Dummy1", "n": 2})
        submission3 = Submission(reddit, _data={"id": "dummy3", "n": 2})
        assert submission1 == submission1
        assert submission2 == submission2
        assert submission3 == submission3
        assert submission1 == submission2
        assert submission2 != submission3
        assert submission1 != submission3
        assert "dummy1" == submission1
        assert submission2 == "dummy1"

    def test_construct_failure(self, reddit):
        message = "Exactly one of 'id', 'url', or '_data' must be provided."
        with pytest.raises(TypeError) as excinfo:
            Submission(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(reddit, id="dummy", url="dummy")
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(reddit, "dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(reddit, url="dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Submission(reddit, "dummy", "dummy", {"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(ValueError):
            Submission(reddit, "")
        with pytest.raises(ValueError):
            Submission(reddit, url="")

    @pytest.mark.filterwarnings("error", category=UserWarning)
    async def test_comment_sort_warning(self, reddit):
        with pytest.raises(UserWarning) as excinfo:
            submission = await reddit.submission("1234", fetch=False)
            submission._fetched = True
            submission.comment_sort = "new"
        assert (
            excinfo.value.args[0]
            == "The comments for this submission have already been fetched, "
            "so the updated comment_sort will not have any effect"
        )

    @pytest.mark.filterwarnings("error", category=UserWarning)
    @pytest.mark.usefixtures("caplog", "reddit")
    async def test_comment_sort_warning__disabled(self, caplog, reddit):
        reddit.config.warn_comment_sort = False
        submission = await reddit.submission("1234", fetch=False)
        submission._fetched = True
        submission.comment_sort = "new"
        assert caplog.records == []

    async def test_comment_unfetched(self, reddit):
        with pytest.raises(TypeError):
            submission = await reddit.submission("1234", fetch=False)
            submission.comments.list()

    def test_construct_from_url(self, reddit):
        assert Submission(reddit, url="http://my.it/2gmzqe") == "2gmzqe"

    def test_fullname(self, reddit):
        submission = Submission(reddit, _data={"id": "dummy"})
        assert submission.fullname == "t3_dummy"

    def test_hash(self, reddit):
        submission1 = Submission(reddit, _data={"id": "dummy1", "n": 1})
        submission2 = Submission(reddit, _data={"id": "Dummy1", "n": 2})
        submission3 = Submission(reddit, _data={"id": "dummy3", "n": 2})
        assert hash(submission1) == hash(submission1)
        assert hash(submission2) == hash(submission2)
        assert hash(submission3) == hash(submission3)
        assert hash(submission1) == hash(submission2)
        assert hash(submission2) != hash(submission3)
        assert hash(submission1) != hash(submission3)

    def test_id_from_url(self):
        urls = [
            "http://my.it/2gmzqe",
            "https://redd.it/2gmzqe",
            "https://redd.it/2gmzqe/",
            "http://reddit.com/comments/2gmzqe",
            "https://www.reddit.com/r/redditdev/comments/2gmzqe/praw_https_enabled_praw_testing_needed/",
            "https://www.reddit.com/gallery/2gmzqe",
        ]
        for url in urls:
            assert Submission.id_from_url(url) == "2gmzqe", url

    def test_id_from_url__invalid_urls(self):
        urls = [
            "",
            "1",
            "/",
            "my.it/2gmzqe",
            "http://my.it/_",
            "https://redd.it/_/",
            "http://reddit.com/comments/_/2gmzqe",
            "https://reddit.com/r/wallpapers/",
            "https://reddit.com/r/wallpapers",
            "https://www.reddit.com/r/test/comments/",
            "https://reddit.com/comments/",
        ]
        for url in urls:
            with pytest.raises(ClientException):
                Submission.id_from_url(url)

    def test_repr(self, reddit):
        submission = Submission(reddit, id="2gmzqe")
        assert repr(submission) == "Submission(id='2gmzqe')"

    def test_str(self, reddit):
        submission = Submission(reddit, _data={"id": "dummy"})
        assert str(submission) == "dummy"

    def test_shortlink(self, reddit):
        submission = Submission(reddit, _data={"id": "dummy"})
        assert submission.shortlink == "https://redd.it/dummy"



"""Test classes from collections.py."""

import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import ClientException, RedditAPIException
from asyncpraw.models import Submission

from ... import IntegrationTest


class TestCollection(IntegrationTest):
    NONEMPTY_REAL_UUID = "3aa31024-711b-46b2-9514-3fd50619f6e8"

    async def test_bad_fetch(self):
        uuid = "A" * 36
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            with pytest.raises(ClientException):
                await subreddit.collections(uuid)

    async def test_init(self):
        uuid = self.NONEMPTY_REAL_UUID
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection1 = await subreddit.collections(uuid)
            collection2 = await subreddit.collections(permalink=collection1.permalink)
            assert collection1 == collection2

    async def test_iter(self):
        uuid = self.NONEMPTY_REAL_UUID
        found_some = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid)
            for post in collection:
                assert isinstance(post, Submission)
                found_some = True
        assert found_some

    async def test_follow(self):
        self.reddit.read_only = False
        uuid = self.NONEMPTY_REAL_UUID
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid)
            await collection.follow()

    async def test_subreddit(self):
        uuid = self.NONEMPTY_REAL_UUID
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid)
            assert str(await collection.subreddit()) in collection.permalink

    async def test_unfollow(self):
        self.reddit.read_only = False
        uuid = self.NONEMPTY_REAL_UUID
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid)
            await collection.unfollow()


class TestCollectionModeration(IntegrationTest):
    NONEMPTY_REAL_UUID = "3aa31024-711b-46b2-9514-3fd50619f6e8"
    UPDATE_LAYOUT_UUID = "3aa31024-711b-46b2-9514-3fd50619f6e8"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_post(self, _):
        self.reddit.read_only = False
        uuid = self.NONEMPTY_REAL_UUID
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid)
            posts = [
                await subreddit.submit(f"Post #{i}", selftext="") for i in range(4)
            ]
            for post in posts:
                await post._fetch()
            # Testing different types for _post_fullname
            await collection.mod.add_post(posts[0])  # Subreddit object
            await collection.mod.add_post(posts[1].fullname)  # fullname
            await collection.mod.add_post(f"https://reddit.com{posts[2].permalink}")
            await collection.mod.add_post(posts[3].id)  # id

            posts.append(
                await subreddit.submit("Post #4", collection_id=uuid, selftext="")
            )

            with pytest.raises(TypeError):
                await collection.mod.add_post(12345)

            await collection._fetch()

            collection_set = set(collection)
            for post in posts:
                assert post in collection_set

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(
                pytest.placeholders.test_subreddit, fetch=True
            )
            collection = await subreddit.collections.mod.create(
                title="Title", description=""
            )
            await collection.mod.delete()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_remove_post(self, _):
        self.reddit.read_only = False
        uuid = self.NONEMPTY_REAL_UUID
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            post = await subreddit.submit("The title", collection_id=uuid, selftext="")
            collection = await subreddit.collections(uuid)
            await collection.mod.remove_post(post)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reorder(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(self.NONEMPTY_REAL_UUID)
            original_order = collection.link_ids
            new_order = (
                collection.link_ids[len(collection.link_ids) // 2 :]
                + collection.link_ids[: len(collection.link_ids) // 2]
            )
            assert len(original_order) == len(new_order)
            assert original_order != new_order
            await collection.mod.reorder(new_order)
            await collection._fetch()
            assert collection.link_ids == new_order

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_description(self, _):
        self.reddit.read_only = False
        uuid = self.NONEMPTY_REAL_UUID
        new_description = "b" * 250
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid)
            await collection.mod.update_description(new_description)
            assert new_description == collection.description

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_display_layout__empty_string(self, _):
        self.reddit.read_only = False
        uuid = self.UPDATE_LAYOUT_UUID
        empty_string = ""
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid, fetch=False)
            await collection.mod.update_display_layout(empty_string)
            await collection.load()
            assert empty_string != collection.display_layout
            assert collection.display_layout is None

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_display_layout__gallery(self, _):
        self.reddit.read_only = False
        uuid = self.UPDATE_LAYOUT_UUID
        gallery_layout = "GALLERY"
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid, fetch=False)
            await collection.mod.update_display_layout(gallery_layout)
            await collection.load()
            assert gallery_layout == collection.display_layout

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_display_layout__invalid_layout(self, _):
        self.reddit.read_only = False
        uuid = self.UPDATE_LAYOUT_UUID
        invalid_layout = "colossal atom cake"
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid, fetch=False)
            with pytest.raises(RedditAPIException):
                await collection.mod.update_display_layout(invalid_layout)
            await collection.load()
            assert collection.display_layout is None

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_display_layout__lowercase(self, _):
        self.reddit.read_only = False
        uuid = self.UPDATE_LAYOUT_UUID
        lowercase_gallery_layout = "gallery"
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid, fetch=False)
            with pytest.raises(RedditAPIException):
                await collection.mod.update_display_layout(lowercase_gallery_layout)
            await collection.load()
            assert collection.display_layout is None

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_display_layout__none(self, _):
        self.reddit.read_only = False
        uuid = self.UPDATE_LAYOUT_UUID
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid, fetch=False)
            await collection.mod.update_display_layout(None)
            await collection.load()
            assert collection.display_layout is None

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_display_layout__timeline(self, _):
        self.reddit.read_only = False
        uuid = self.UPDATE_LAYOUT_UUID
        timeline_layout = "TIMELINE"
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid, fetch=False)
            await collection.mod.update_display_layout(timeline_layout)
            await collection.load()
            assert timeline_layout == collection.display_layout

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_title(self, _):
        self.reddit.read_only = False
        uuid = self.NONEMPTY_REAL_UUID
        new_title = "a" * 100
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections(uuid)
            await collection.mod.update_title(new_title)
            assert new_title == collection.title


class TestSubredditCollections(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_call(self, _):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = next(iter(await self.async_list(subreddit.collections)))
            test_collection = await subreddit.collections(collection.collection_id)
            assert collection == test_collection
            test_collection = await subreddit.collections(
                permalink=collection.permalink
            )
            assert collection == test_collection

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_iter(self, _):
        with self.use_cassette():
            found_any = False
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for collection in subreddit.collections:
                assert collection.permalink
                assert collection.title is not None
                assert collection.description is not None
                found_any = True
            assert found_any


class TestSubredditCollectionsModeration(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create(self, _):
        title = "The title!"
        description = "The description."
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections.mod.create(
                title=title, description=description
            )
            assert collection.title == title
            assert collection.description == description
            assert len(collection) == 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create__empty_layout(self, _):
        title = "The title!"
        description = "The description."
        layout = ""
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections.mod.create(
                title=title, description=description, display_layout=layout
            )
            assert collection.title == title
            assert collection.description == description
            assert collection.display_layout is None
            assert len(collection) == 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create__gallery_layout(self, _):
        title = "The title!"
        description = "The description."
        layout = "GALLERY"
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections.mod.create(
                title=title, description=description, display_layout=layout
            )
            assert collection.title == title
            assert collection.description == description
            assert collection.display_layout == layout
            assert len(collection) == 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create__invalid_layout(self, _):
        title = "The title!"
        description = "The description."
        layout = "milk before cereal"
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException):
                subreddit = await self.reddit.subreddit(
                    pytest.placeholders.test_subreddit
                )
                await subreddit.collections.mod.create(
                    title=title, description=description, display_layout=layout
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create__lowercase_layout(self, _):
        title = "The title!"
        description = "The description."
        layout = "gallery"
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException):
                subreddit = await self.reddit.subreddit(
                    pytest.placeholders.test_subreddit
                )
                await subreddit.collections.mod.create(
                    title=title, description=description, display_layout=layout
                )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create__none_layout(self, _):
        title = "The title!"
        description = "The description."
        layout = None
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections.mod.create(
                title=title, description=description, display_layout=layout
            )
            assert collection.title == title
            assert collection.description == description
            assert collection.display_layout is None
            assert len(collection) == 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create__timeline_layout(self, _):
        title = "The title!"
        description = "The description."
        layout = "TIMELINE"
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            collection = await subreddit.collections.mod.create(
                title=title, description=description, display_layout=layout
            )
            assert collection.title == title
            assert collection.description == description
            assert collection.display_layout == layout
            assert len(collection) == 0



"""Provide the ModAction class."""
from typing import TYPE_CHECKING, Union

from .base import AsyncPRAWBase
from .reddit.redditor import Redditor

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class ModAction(AsyncPRAWBase):
    """Represent a moderator action."""

    @property
    def mod(self) -> "asyncpraw.models.Redditor":
        """Return the :class:`.Redditor` who the action was issued by."""
        return Redditor(self._reddit, name=self._mod)  # pylint: disable=no-member

    @mod.setter
    def mod(self, value: Union[str, "asyncpraw.models.Redditor"]):
        self._mod = value  # pylint: disable=attribute-defined-outside-init



"""Test asyncpraw.objector."""
import pytest

from asyncpraw.exceptions import RedditAPIException
from asyncpraw.models import Submission

from . import IntegrationTest


class TestObjector(IntegrationTest):
    async def test_raise_api_exception(self):
        message = "USER_REQUIRED: 'Please log in to do that.'"
        with self.use_cassette():
            submission = Submission(self.reddit, "4b536h")
            with pytest.raises(RedditAPIException) as excinfo:
                await submission.mod.approve()
            assert excinfo.value.items[0].error_type == "USER_REQUIRED"
            assert str(excinfo.value.items[0]) == message



"""Provide the Front class."""
from typing import TYPE_CHECKING, AsyncIterator, Union
from urllib.parse import urljoin

from .listing.generator import ListingGenerator
from .listing.mixins import SubredditListingMixin

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class Front(SubredditListingMixin):
    """Front is a Listing class that represents the front page."""

    def __init__(self, reddit: "asyncpraw.Reddit"):
        """Initialize a :class:`.Front` instance."""
        super().__init__(reddit, _data=None)
        self._path = "/"

    def best(
        self, **generator_kwargs: Union[str, int]
    ) -> AsyncIterator["asyncpraw.models.Submission"]:
        """Return a :class:`.ListingGenerator` for best items.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        """
        return ListingGenerator(
            self._reddit, urljoin(self._path, "best"), **generator_kwargs
        )
